# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Meshing workflow specialization of the Workflow module that wraps and extends the
core functionality."""

from __future__ import annotations

from enum import Enum

from ansys.fluent.core.services.datamodel_se import PyMenuGeneric
from ansys.fluent.core.utils.fluent_version import FluentVersion
from ansys.fluent.core.workflow import Workflow

name_to_identifier_map = {
    "Watertight Geometry": "EnableCleanCAD",
    "Fault-tolerant Meshing": "EnableComplexMeshing",
    "2D Meshing": "EnablePrime2dMeshing",
    "Topology Based Meshing": "EnablePrimeMeshing",
}


class MeshingWorkflow(Workflow):
    """Provides meshing specialization of the workflow wrapper that extends the core
    functionality in an object-oriented manner."""

    def __init__(
        self,
        workflow: PyMenuGeneric,
        meshing: PyMenuGeneric,
        name: str,
        identifier: str,
        fluent_version: FluentVersion,
        initialize: bool = True,
    ) -> None:
        """Initialize MeshingWorkflow.

        Parameters
        ----------
        workflow : PyMenuGeneric
            Underlying workflow object.
        meshing : PyMenuGeneric
            Meshing object.
        name: str
            Workflow name to initialize it.
        identifier: str
            Workflow name to identify it from global settings.
        fluent_version: FluentVersion
            Version of Fluent in this session.
        initialize: bool
            Flag to initialize the workflow, defaults to True.
        """
        super().__init__(
            workflow=workflow, command_source=meshing, fluent_version=fluent_version
        )
        self._meshing = meshing
        self._name = name
        self._identifier = identifier
        self._unsubscribe_root_affected_callback()
        if initialize:
            self._new_workflow(name=self._name)
        else:
            self._activate_dynamic_interface(dynamic_interface=True)

    def __getattribute__(self, item: str):
        if (
            not item.startswith("_")
            and not getattr(self._meshing.GlobalSettings, self._identifier)()
        ):
            raise RuntimeError(
                f"'{self._name}' objects are inaccessible from other workflows."
            )
        return super().__getattribute__(item)


class WatertightMeshingWorkflow(MeshingWorkflow):
    """Provides watertight meshing specialization of the workflow wrapper."""

    def __init__(
        self,
        workflow: PyMenuGeneric,
        meshing: PyMenuGeneric,
        fluent_version: FluentVersion,
        initialize: bool = True,
    ) -> None:
        """Initialize WatertightMeshingWorkflow.

        Parameters
        ----------
        workflow : PyMenuGeneric
            Underlying workflow object.
        meshing : PyMenuGeneric
            Meshing object.
        fluent_version: FluentVersion
            Version of Fluent in this session.
        initialize: bool
            Flag to initialize the workflow, defaults to True.
        """
        super().__init__(
            workflow=workflow,
            meshing=meshing,
            name="Watertight Geometry",
            identifier=name_to_identifier_map["Watertight Geometry"],
            fluent_version=fluent_version,
            initialize=initialize,
        )


class FaultTolerantMeshingWorkflow(MeshingWorkflow):
    """Provides fault-tolerant meshing specialization of the workflow wrapper."""

    def __init__(
        self,
        workflow: PyMenuGeneric,
        meshing: PyMenuGeneric,
        part_management: PyMenuGeneric,
        pm_file_management: PyMenuGeneric,
        fluent_version: FluentVersion,
        initialize: bool = True,
    ) -> None:
        """Initialize FaultTolerantMeshingWorkflow.

        Parameters
        ----------
        workflow : PyMenuGeneric
            Underlying workflow object.
        meshing : PyMenuGeneric
            Meshing object.
        part_management : PyMenuGeneric
            Part management object.
        pm_file_management : PyMenuGeneric
            File management object in the part management object.
        fluent_version: FluentVersion
            Version of Fluent in this session.
        initialize: bool
            Flag to initialize the workflow, defaults to True.
        """
        super().__init__(
            workflow=workflow,
            meshing=meshing,
            name="Fault-tolerant Meshing",
            identifier=name_to_identifier_map["Fault-tolerant Meshing"],
            fluent_version=fluent_version,
            initialize=initialize,
        )
        self._part_management = part_management
        self._pm_file_management = pm_file_management

    @property
    def part_management(self) -> PyMenuGeneric | None:
        """Access part-management in fault-tolerant mode.

        Returns
        -------
        PyMenuGeneric | None
            Part-management.
        """
        return self._part_management

    @property
    def pm_file_management(self):
        """Access the part-management file-management object in fault-tolerant mode.

        Returns
        -------
        PyMenuGeneric | None
            File management object in the part management object.
        """
        return self._pm_file_management


class TwoDimensionalMeshingWorkflow(MeshingWorkflow):
    """Provides 2D meshing specialization of the workflow wrapper."""

    def __init__(
        self,
        workflow: PyMenuGeneric,
        meshing: PyMenuGeneric,
        fluent_version: FluentVersion,
        initialize: bool = True,
    ) -> None:
        """Initialize TwoDimensionalMeshingWorkflow.

        Parameters
        ----------
        workflow : PyMenuGeneric
            Underlying workflow object.
        meshing : PyMenuGeneric
            Meshing object.
        fluent_version: FluentVersion
            Version of Fluent in this session.
        initialize: bool
            Flag to initialize the workflow, defaults to True.
        """
        super().__init__(
            workflow=workflow,
            meshing=meshing,
            name="2D Meshing",
            identifier=name_to_identifier_map["2D Meshing"],
            fluent_version=fluent_version,
            initialize=initialize,
        )


class TopologyBasedMeshingWorkflow(MeshingWorkflow):
    """Provides topology-based meshing specialization of the workflow wrapper."""

    def __init__(
        self,
        workflow: PyMenuGeneric,
        meshing: PyMenuGeneric,
        fluent_version: FluentVersion,
        initialize: bool = True,
    ) -> None:
        """Initialize TopologyBasedMeshingWorkflow.

        Parameters
        ----------
        workflow : PyMenuGeneric
            Underlying workflow object.
        meshing : PyMenuGeneric
            Meshing object.
        fluent_version: FluentVersion
            Version of Fluent in this session.
        initialize: bool
            Flag to initialize the workflow, defaults to True.
        """
        super().__init__(
            workflow=workflow,
            meshing=meshing,
            name="Topology Based Meshing",
            identifier=name_to_identifier_map["Topology Based Meshing"],
            fluent_version=fluent_version,
            initialize=initialize,
        )


class WorkflowMode(Enum):
    """Provides an enum of supported Fluent meshing workflow modes."""

    WATERTIGHT_MESHING_MODE = WatertightMeshingWorkflow
    FAULT_TOLERANT_MESHING_MODE = FaultTolerantMeshingWorkflow
    TWO_DIMENSIONAL_MESHING_MODE = TwoDimensionalMeshingWorkflow
    TOPOLOGY_BASED_MESHING_MODE = TopologyBasedMeshingWorkflow


class LoadWorkflow(Workflow):
    """Provides a specialization of the workflow wrapper for a loaded workflow."""

    def __init__(
        self,
        workflow: PyMenuGeneric,
        meshing: PyMenuGeneric,
        file_path: str,
        fluent_version: FluentVersion,
    ) -> None:
        """Initialize a ``LoadWorkflow`` instance.

        Parameters
        ----------
        workflow : PyMenuGeneric
            Underlying workflow object.
        meshing : PyMenuGeneric
            Meshing object.
        file_path: str
            Path to the saved workflow.
        fluent_version: FluentVersion
            Version of Fluent in this session.
        """
        super().__init__(
            workflow=workflow, command_source=meshing, fluent_version=fluent_version
        )
        self._meshing = meshing
        self._unsubscribe_root_affected_callback()
        self._load_workflow(file_path=file_path)


class CreateWorkflow(Workflow):
    """Provides a specialization of the workflow wrapper for a newly created
    workflow."""

    def __init__(
        self,
        workflow: PyMenuGeneric,
        meshing: PyMenuGeneric,
        fluent_version: FluentVersion,
        initialize: bool = True,
    ) -> None:
        """Initialize a ``CreateWorkflow`` instance.

        Parameters
        ----------
        workflow : PyMenuGeneric
            Underlying workflow object.
        meshing : PyMenuGeneric
            Meshing object.
        fluent_version: FluentVersion
            Version of Fluent in this session.
        initialize: bool
            Flag to initialize the workflow, defaults to True.
        """
        super().__init__(
            workflow=workflow, command_source=meshing, fluent_version=fluent_version
        )
        self._meshing = meshing
        self._unsubscribe_root_affected_callback()
        if initialize:
            self._create_workflow()
        else:
            self._activate_dynamic_interface(dynamic_interface=True)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
import pytest

from ansys.fluent.core import examples
from ansys.fluent.core.examples.downloads import download_file


@pytest.mark.fluent_version(">=23.2")
def test_solution_variables(new_solver_session):
    solver = new_solver_session
    import_file_name = examples.download_file(
        "mixing_elbow.cas.h5", "pyfluent/mixing_elbow"
    )
    examples.download_file("mixing_elbow.dat.h5", "pyfluent/mixing_elbow")

    solution_variable_info = solver.fields.solution_variable_info
    solution_variable_data = solver.fields.solution_variable_data
    assert solution_variable_info == solver.svar_info
    assert solution_variable_data == solver.svar_data

    solver.file.read_case_data(file_name=import_file_name)

    zones_info = solution_variable_info.get_zones_info()

    assert zones_info.domains == ["mixture"]

    assert set(zones_info.zones) == {
        "symmetry-xyplane",
        "hot-inlet",
        "cold-inlet",
        "outlet",
        "wall-inlet",
        "wall-elbow",
        "elbow-fluid",
        "interior--elbow-fluid",
    }

    zone_info = zones_info["wall-inlet"]

    assert zone_info.name == "wall-inlet"

    assert zone_info.count == 268

    assert zone_info.zone_id == 33

    assert zone_info.zone_type == "wall"

    wall_fluid_info = solution_variable_info.get_variables_info(
        zone_names=["wall-elbow", "elbow-fluid"], domain_name="mixture"
    )

    assert set(wall_fluid_info.solution_variables) == {
        "SV_ADS_0",
        "SV_ADS_1",
        "SV_CENTROID",
        "SV_H",
        "SV_K",
        "SV_O",
        "SV_P",
        "SV_T",
        "SV_U",
        "SV_V",
        "SV_W",
    }

    solution_variable_info_centroid = wall_fluid_info["SV_CENTROID"]

    assert solution_variable_info_centroid.name == "SV_CENTROID"

    assert solution_variable_info_centroid.dimension == 3

    assert solution_variable_info_centroid.field_type == np.float64

    sv_p_wall_fluid = solution_variable_data.get_data(
        solution_variable_name="SV_P",
        zone_names=["elbow-fluid", "wall-elbow"],
        domain_name="mixture",
    )
    assert sv_p_wall_fluid.domain == "mixture"

    assert sv_p_wall_fluid.zones == ["wall-elbow", "elbow-fluid"]

    fluid_temp = sv_p_wall_fluid["elbow-fluid"]
    assert fluid_temp.size == 17822
    assert str(fluid_temp.dtype) == "float64"

    wall_press_array = solution_variable_data.create_empty_array(
        "SV_P", "wall-elbow", "mixture"
    )
    fluid_press_array = solution_variable_data.create_empty_array(
        "SV_P", "elbow-fluid", "mixture"
    )
    wall_press_array[:] = 500
    fluid_press_array[:] = 600
    zone_names_to_solution_variable_data = {
        "wall-elbow": wall_press_array,
        "elbow-fluid": fluid_press_array,
    }
    solution_variable_data.set_data(
        solution_variable_name="SV_P",
        zone_names_to_solution_variable_data=zone_names_to_solution_variable_data,
        domain_name="mixture",
    )

    updated_sv_p_data = solution_variable_data.get_data(
        solution_variable_name="SV_P",
        zone_names=["elbow-fluid", "wall-elbow"],
        domain_name="mixture",
    )

    assert updated_sv_p_data.domain == "mixture"
    assert updated_sv_p_data.zones == ["wall-elbow", "elbow-fluid"]

    assert updated_sv_p_data["elbow-fluid"].size == 17822
    assert str(updated_sv_p_data["elbow-fluid"].dtype) == "float64"

    assert updated_sv_p_data["wall-elbow"][0] == 500.0
    assert updated_sv_p_data["elbow-fluid"][-1] == 600.0


@pytest.mark.fluent_version(">=23.2")
def test_solution_variables_single_precision(new_solver_session_sp):
    solver = new_solver_session_sp
    import_file_name = examples.download_file(
        "mixing_elbow.cas.h5", "pyfluent/mixing_elbow"
    )
    examples.download_file("mixing_elbow.dat.h5", "pyfluent/mixing_elbow")

    solution_variable_info = solver.fields.solution_variable_info
    solution_variable_data = solver.fields.solution_variable_data
    assert solution_variable_info == solver.svar_info
    assert solution_variable_data == solver.svar_data

    solver.file.read_case_data(file_name=import_file_name)

    zones_info = solution_variable_info.get_zones_info()

    assert zones_info.domains == ["mixture"]

    assert set(zones_info.zones) == {
        "symmetry-xyplane",
        "hot-inlet",
        "cold-inlet",
        "outlet",
        "wall-inlet",
        "wall-elbow",
        "elbow-fluid",
        "interior--elbow-fluid",
    }

    zone_info = zones_info["wall-elbow"]

    assert zone_info.name == "wall-elbow"

    assert zone_info.count == 2168

    assert zone_info.zone_id == 34

    assert zone_info.zone_type == "wall"

    wall_fluid_info = solution_variable_info.get_variables_info(
        zone_names=["wall-elbow", "elbow-fluid"], domain_name="mixture"
    )

    assert set(wall_fluid_info.solution_variables) == {
        "SV_ADS_0",
        "SV_ADS_1",
        "SV_CENTROID",
        "SV_H",
        "SV_K",
        "SV_O",
        "SV_P",
        "SV_T",
        "SV_U",
        "SV_V",
        "SV_W",
    }

    solution_variable_info_centroid = wall_fluid_info["SV_CENTROID"]

    assert solution_variable_info_centroid.name == "SV_CENTROID"

    assert solution_variable_info_centroid.dimension == 3

    assert solution_variable_info_centroid.field_type == np.float32

    sv_p_wall_fluid = solution_variable_data.get_data(
        solution_variable_name="SV_P",
        zone_names=["wall-elbow", "elbow-fluid"],
        domain_name="mixture",
    )
    assert sv_p_wall_fluid.domain == "mixture"

    assert sv_p_wall_fluid.zones == ["wall-elbow", "elbow-fluid"]

    fluid_temp = sv_p_wall_fluid["elbow-fluid"]
    assert fluid_temp.size == 17822
    assert str(fluid_temp.dtype) == "float32"


@pytest.mark.fluent_version(">=24.2")
def test_solution_variable_does_not_modify_case(new_solver_session):
    solver = new_solver_session
    case_path = download_file("mixing_elbow.cas.h5", "pyfluent/mixing_elbow")
    download_file("mixing_elbow.dat.h5", "pyfluent/mixing_elbow")
    solver.file.read_case_data(file_name=case_path)
    solver.scheme_eval.scheme_eval("(%save-case-id)")
    assert not solver.scheme_eval.scheme_eval("(case-modified?)")
    solver.fields.solution_variable_data.get_data(
        solution_variable_name="SV_P",
        zone_names=["elbow-fluid", "wall-elbow"],
        domain_name="mixture",
    )
    assert not solver.scheme_eval.scheme_eval("(case-modified?)")


@pytest.mark.fluent_version(">=25.2")
def test_solution_variable_udm_data(mixing_elbow_case_session_t4):
    solver = mixing_elbow_case_session_t4
    solver.tui.define.user_defined.user_defined_memory("2")
    solver.settings.solution.initialization.hybrid_initialize()
    solver.settings.solution.run_calculation.iterate(iter_count=1)
    udm_data = solver.fields.solution_variable_data.get_data(
        solution_variable_name="SV_UDM_I",
        domain_name="mixture",
        zone_names=["wall-elbow"],
    )["wall-elbow"]
    np.testing.assert_array_equal(udm_data, np.zeros(4336))
    udm_data[:2168] = 5
    udm_data[2168:] = 10
    solver.fields.solution_variable_data.set_data(
        solution_variable_name="SV_UDM_I",
        domain_name="mixture",
        zone_names_to_solution_variable_data={"wall-elbow": udm_data},
    )
    new_array = solver.fields.solution_variable_data.get_data(
        solution_variable_name="SV_UDM_I",
        domain_name="mixture",
        zone_names=["wall-elbow"],
    )["wall-elbow"]
    np.testing.assert_array_equal(new_array, udm_data)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Test the PyPIM integration."""

from concurrent import futures
import os
from pathlib import Path
import shutil
from unittest.mock import create_autospec
import uuid

import grpc
from grpc_health.v1 import health_pb2_grpc
import pytest
from test_session import MockHealthServicer, MockSchemeEvalServicer

from ansys.api.fluent.v0 import scheme_eval_pb2_grpc
import ansys.fluent.core as pyfluent
from ansys.fluent.core import EXAMPLES_PATH, examples
from ansys.fluent.core.fluent_connection import (
    FluentConnection,
    UnsupportedRemoteFluentInstance,
)
from ansys.fluent.core.launcher import launcher
from ansys.fluent.core.session import BaseSession
from ansys.fluent.core.session_pure_meshing import PureMeshing
from ansys.fluent.core.session_solver import Solver
from ansys.fluent.core.utils.file_transfer_service import PimFileTransferService
import ansys.fluent.core.utils.fluent_version as docker_image_version
from ansys.fluent.core.utils.fluent_version import FluentVersion
from ansys.fluent.core.utils.networking import get_free_port
import ansys.platform.instancemanagement as pypim


def test_launch_remote_instance(monkeypatch, new_solver_session):
    monkeypatch.setattr(pyfluent, "CHECK_HEALTH", False)
    fluent = new_solver_session
    # Create a mock pypim pretending it is configured and returning a channel to an already running Fluent
    mock_instance = pypim.Instance(
        definition_name="definitions/fake-fluent",
        name="instances/fake-fluent",
        ready=True,
        status_message=None,
        services={
            "grpc": pypim.Service(
                uri=fluent._fluent_connection._channel_str, headers={}
            )
        },
    )
    pim_channel = grpc.insecure_channel(
        fluent._fluent_connection._channel_str,
    )
    mock_instance.wait_for_ready = create_autospec(mock_instance.wait_for_ready)
    mock_instance.build_grpc_channel = create_autospec(
        mock_instance.build_grpc_channel, return_value=pim_channel
    )
    mock_instance.delete = create_autospec(mock_instance.delete)

    mock_client = pypim.Client(channel=grpc.insecure_channel("localhost:12345"))
    mock_client.create_instance = create_autospec(
        mock_client.create_instance, return_value=mock_instance
    )

    mock_connect = create_autospec(pypim.connect, return_value=mock_client)
    mock_is_configured = create_autospec(pypim.is_configured, return_value=True)
    monkeypatch.setattr(pypim, "connect", mock_connect)
    monkeypatch.setattr(pypim, "is_configured", mock_is_configured)

    if os.getenv("FLUENT_IMAGE_TAG"):
        monkeypatch.setattr(
            FluentVersion,
            "get_latest_installed",
            lambda: docker_image_version.get_version(),
        )

    # Start fluent with launch_fluent
    # Note: This is mocking to start Fluent, but actually reusing the common one
    # Thus cleanup_on_exit is set to false
    fluent = launcher.launch_fluent(cleanup_on_exit=False, mode="solver")

    # Assert: PyFluent went through the pypim workflow
    assert mock_is_configured.called
    assert mock_connect.called

    mock_client.create_instance.assert_called_with("fluent-3ddp", product_version=None)
    assert mock_instance.wait_for_ready.called
    mock_instance.build_grpc_channel.assert_called_with()

    # And it connected using the channel created by PyPIM
    assert fluent._fluent_connection._channel == pim_channel

    # and it kept track of the instance to be able to delete it
    assert fluent._fluent_connection._remote_instance == mock_instance

    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    ip = "127.0.0.1"
    port = get_free_port()
    server.add_insecure_port(f"{ip}:{port}")
    health_pb2_grpc.add_HealthServicer_to_server(MockHealthServicer(), server)
    scheme_eval_pb2_grpc.add_SchemeEvalServicer_to_server(
        MockSchemeEvalServicer(), server
    )
    server.start()

    with pytest.raises(UnsupportedRemoteFluentInstance):
        fluent_connection = FluentConnection(
            ip=ip,
            port=port,
            password="12345",
            remote_instance=mock_instance,
            cleanup_on_exit=False,
        )
        session = BaseSession(
            fluent_connection=fluent_connection,
            scheme_eval=fluent_connection._connection_interface.scheme_eval,
        )

        file_transfer_service = PimFileTransferService(pim_instance=mock_instance)
        assert not file_transfer_service.file_service
        assert file_transfer_service.is_configured()
        assert file_transfer_service.pim_instance
        assert file_transfer_service.upload_server

        session.exit(wait=60)
        session._fluent_connection.wait_process_finished(wait=60)


class TransferRequestRecorder:
    def __init__(self):
        self.uploaded_files = list()
        self.downloaded_files = list()

    def uploads(self):
        return self.uploaded_files

    def downloads(self):
        return self.downloaded_files

    def upload(self, file_name: str):
        self.uploaded_files.append(file_name)

    def download(self, file_name: str):
        self.downloaded_files.append(file_name)

    def is_configured(self):
        return True


def rename_downloaded_file(file_path: str, suffix: str) -> str:
    """Rename downloaded file by appending a suffix to the file name.

    Parameters
    ----------
    file_path : str
        Downloaded file path. Can be absolute or relative.
    suffix : str
        Suffix to append to the file name.

    Returns:
    --------
    str
        New file path with the suffix appended to the file name.
    """
    ext = "".join(Path(file_path).suffixes)
    orig_path = Path(file_path)
    file_path = file_path.removesuffix(ext)
    file_path = Path(file_path)
    if file_path.is_absolute():
        new_stem = f"{file_path.stem}{suffix}"
        new_path = file_path.with_stem(new_stem)
        new_path = new_path.with_suffix(ext)
        orig_path.rename(new_path)
        return str(new_path)
    else:
        orig_abs_path = Path(EXAMPLES_PATH) / orig_path
        abs_path = Path(EXAMPLES_PATH) / file_path
        new_stem = f"{file_path.stem}{suffix}"
        new_path = abs_path.with_stem(new_stem)
        new_path = new_path.with_suffix(ext)
        orig_abs_path.rename(new_path)
        return str(file_path.with_stem(new_stem).with_suffix(ext))


@pytest.mark.parametrize(
    "ext,a,b,c,d",
    [(".cas", "a1", "b1", "c1", "d1"), (".cas.gz", "a2", "b2", "c2", "d2")],
)
def test_rename_downloaded_file(ext, a, b, c, d):
    try:
        file_path = Path(EXAMPLES_PATH) / f"{a}{ext}"
        file_path.touch()
        file_path = str(file_path)
        new_file_path = rename_downloaded_file(file_path, "_1")
        assert new_file_path == str(Path(EXAMPLES_PATH) / f"{a}_1{ext}")
    except Exception:
        raise
    finally:
        Path(new_file_path).unlink(missing_ok=True)

    try:
        file_path = f"{b}{ext}"
        (Path(EXAMPLES_PATH) / file_path).touch()
        new_file_path = rename_downloaded_file(file_path, "_1")
        assert new_file_path == f"{b}_1{ext}"
    except Exception:
        raise
    finally:
        (Path(EXAMPLES_PATH) / new_file_path).unlink(missing_ok=True)

    try:
        dir_path = Path(EXAMPLES_PATH) / c
        dir_path.mkdir()
        file_path = dir_path / f"{d}{ext}"
        file_path.touch()
        file_path = str(Path(c) / f"{d}{ext}")
        new_file_path = rename_downloaded_file(file_path, "_1")
        assert new_file_path == str(Path(c) / f"{d}_1{ext}")
    except Exception:
        raise
    finally:
        shutil.rmtree(dir_path, ignore_errors=True)


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.2")
def test_file_purpose_on_remote_instance(
    monkeypatch, new_solver_session, new_meshing_session
):
    solver = new_solver_session

    file_service = TransferRequestRecorder()

    solver_session = Solver(
        fluent_connection=solver._fluent_connection,
        scheme_eval=solver._fluent_connection._connection_interface.scheme_eval,
        file_transfer_service=file_service,
    )

    import_file_name = examples.download_file(
        "mixing_elbow.msh.h5", "pyfluent/mixing_elbow"
    )
    suffix = uuid.uuid4().hex
    import_file_name = rename_downloaded_file(import_file_name, f"_{suffix}")

    solver_session.file.read_case(file_name=import_file_name)
    assert file_service.is_configured()
    assert file_service.uploads()
    assert len(file_service.uploads()) == 1
    assert file_service.uploads()[0] == import_file_name

    solver_session.file.write_case(file_name=import_file_name)
    assert file_service.downloads()
    assert len(file_service.downloads()) == 1
    assert file_service.downloads()[0] == import_file_name

    solver_session.exit()

    meshing = new_meshing_session

    meshing_session = PureMeshing(
        fluent_connection=meshing._fluent_connection,
        scheme_eval=meshing._fluent_connection._connection_interface.scheme_eval,
        file_transfer_service=file_service,
    )

    meshing_session.meshing.File.ReadMesh(FileName=import_file_name)
    assert len(file_service.uploads()) == 2
    assert file_service.uploads()[1] == import_file_name

    meshing_session.meshing.File.WriteMesh(FileName=import_file_name)
    assert len(file_service.downloads()) == 2
    assert file_service.downloads()[1] == import_file_name

    meshing_session.exit()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provide a module to generate explicit Fluent TUI menu classes.

This module starts up Fluent and calls the underlying gRPC APIs to generate the
following TUI Python modules:

- src/ansys/fluent/core/solver/tui.py
- src/ansys/fluent/core/meshing/tui.py.

Usage
-----

`python codegen/tuigen.py`
"""

import argparse
import logging
import os
from pathlib import Path
import pickle
import platform
import shutil
import string
import subprocess
from typing import Any, Dict
import uuid

from defusedxml.ElementTree import parse

import ansys.fluent.core as pyfluent
from ansys.fluent.core import FluentMode, launch_fluent
from ansys.fluent.core.codegen import StaticInfoType
from ansys.fluent.core.codegen.data.fluent_gui_help_patch import XML_HELP_PATCH
from ansys.fluent.core.services.datamodel_tui import (
    convert_path_to_grpc_path,
    convert_tui_menu_to_func_name,
)
from ansys.fluent.core.utils.fix_doc import escape_wildcards
from ansys.fluent.core.utils.fluent_version import (
    AnsysVersionNotFound,
    FluentVersion,
    get_version_for_file_name,
)

logger = logging.getLogger("pyfluent.tui")

_ROOT_DIR = Path(__file__) / ".." / ".." / ".." / ".." / ".." / ".."


def _get_tui_filepath(mode: str, version: str):
    return (pyfluent.CODEGEN_OUTDIR / mode / f"tui_{version}.py").resolve()


_INDENT_STEP = 4


# TODO: Move doc-specific variables to docgen


def _get_tui_docdir(mode: str):
    return os.path.normpath(
        os.path.join(
            _ROOT_DIR,
            "doc",
            "source",
            "api",
            mode,
            "tui",
        )
    )


_XML_HELP_FILE = (Path(__file__) / ".." / "data" / "fluent_gui_help.xml").resolve()
_XML_HELPSTRINGS = {}


def _copy_tui_help_xml_file(version: str):
    if os.getenv("PYFLUENT_LAUNCH_CONTAINER") == "1":
        image_tag = os.getenv("FLUENT_IMAGE_TAG", "v25.1.0")
        image_name = f"ghcr.io/ansys/pyfluent:{image_tag}"
        container_name = uuid.uuid4().hex
        is_linux = platform.system() == "Linux"
        subprocess.run(
            f"docker container create --name {container_name} {image_name}",
            shell=is_linux,
        )
        xml_source = f"/ansys_inc/v{version}/commonfiles/help/en-us/fluent_gui_help/fluent_gui_help.xml"
        subprocess.run(
            f"docker cp {container_name}:{xml_source} {str(_XML_HELP_FILE)}",
            shell=is_linux,
        )
        subprocess.run(f"docker container rm {container_name}", shell=is_linux)

    else:
        try:
            ansys_version = (
                FluentVersion.get_latest_installed()
            )  # picking up the file from the latest install location
            awp_root = os.environ[ansys_version.awp_var]
            xml_source = (
                Path(awp_root)
                / "commonfiles"
                / "help"
                / "en-us"
                / "fluent_gui_help"
                / "fluent_gui_help.xml"
            )
            if xml_source.exists():
                shutil.copy(str(xml_source), _XML_HELP_FILE)
            else:
                logger.warning("fluent_gui_help.xml is not found.")
        except AnsysVersionNotFound:
            logger.warning("fluent_gui_help.xml is not found.")


def _populate_xml_helpstrings():
    if not Path(_XML_HELP_FILE).exists():
        return

    tree = parse(_XML_HELP_FILE)
    root = tree.getroot()
    help_contents_node = root.find(".//*[@id='flu_tui_help_contents']")
    field_help_node = help_contents_node.find(".//*[@id='fluent_tui_field_help']")

    for node in field_help_node.findall("sect2"):
        id = node.get("id")
        k = node.find("h3").text
        k = k.strip().strip("/")
        path = k.split("/")
        path = [c.rstrip("?").replace("-", "_") for c in path]
        k = "/" + "/".join(path)
        patched_doc = XML_HELP_PATCH.get(id)
        if patched_doc:
            _XML_HELPSTRINGS[k] = patched_doc
        else:
            v = "".join(node.find("p").itertext())
            _XML_HELPSTRINGS[k] = v
    _XML_HELP_FILE.unlink(missing_ok=True)


def _is_valid_tui_menu_name(name):
    return name and not all(x in string.punctuation for x in name)


class _TUIMenu:
    """Class representing Fluent's TUI menu."""

    def __init__(self, path: str, doc: str, is_command: bool = False):
        self.path = path
        self.tui_name = path[-1] if path else ""
        self.name = convert_tui_menu_to_func_name(self.tui_name)
        self.is_command = is_command
        tui_path = convert_path_to_grpc_path(path)
        self.doc = _XML_HELPSTRINGS.get(tui_path, None)
        if self.doc:
            del _XML_HELPSTRINGS[tui_path]
        else:
            self.doc = doc
        self.doc = escape_wildcards(self.doc)
        self.doc = self.doc.strip()
        if not self.doc.endswith("."):
            self.doc = self.doc + "."
        self.children = {}

    def get_command_path(self, command: str) -> str:
        """Get the full path to a command."""
        return convert_path_to_grpc_path(self.path + [command])


class _RenameModuleUnpickler(pickle.Unpickler):
    def find_class(self, module, name):
        renamed_module = module
        if module == "tuigen":
            renamed_module = "ansys.fluent.core.codegen.tuigen"

        return super(_RenameModuleUnpickler, self).find_class(renamed_module, name)


class TUIGenerator:
    """Generates explicit TUI menu classes."""

    def __init__(
        self, mode: str, version: str, static_infos: dict, verbose: bool = False
    ):
        self._mode = mode
        self._version = version
        self._tui_file = _get_tui_filepath(mode, version)
        if Path(self._tui_file).exists():
            Path(self._tui_file).unlink()
        self._tui_doc_dir = _get_tui_docdir(mode)
        self._tui_heading = mode + ".tui"
        self._tui_module = "ansys.fluent.core." + self._tui_heading + f"_{version}"
        if Path(self._tui_doc_dir).exists():
            shutil.rmtree(Path(self._tui_doc_dir))
        self._main_menu = _TUIMenu([], "")
        self._static_infos = static_infos
        self._verbose = verbose

    def _populate_menu(self, menu: _TUIMenu, info: Dict[str, Any]):
        for child_menu_name, child_menu_info in sorted(info["menus"].items()):
            if _is_valid_tui_menu_name(child_menu_name):
                child_menu = _TUIMenu(
                    menu.path + [child_menu_name], child_menu_info["help"]
                )
                menu.children[child_menu.name] = child_menu
                self._populate_menu(child_menu, child_menu_info)
        for child_command_name, child_command_info in sorted(info["commands"].items()):
            if _is_valid_tui_menu_name(child_command_name):
                child_menu = _TUIMenu(
                    menu.path + [child_command_name], child_command_info["help"], True
                )
                menu.children[child_menu.name] = child_menu

    def _write_code_to_tui_file(self, code: str, indent: int = 0):
        self.__writer.write(" " * _INDENT_STEP * indent + code)

    def _write_menu_to_tui_file(self, menu: _TUIMenu, indent: int = 0):
        api_tree = {}
        self._write_code_to_tui_file("\n")
        self._write_code_to_tui_file(f"class {menu.name}(TUIMenu):\n", indent)
        indent += 1
        self._write_code_to_tui_file('"""\n', indent)
        doc_lines = menu.doc.splitlines()
        for line in doc_lines:
            line = line.strip()
            if line:
                self._write_code_to_tui_file(f"{line}\n", indent)
        self._write_code_to_tui_file('"""\n', indent)
        self._write_code_to_tui_file(
            "def __init__(self, service, version, mode, path):\n", indent
        )
        indent += 1
        for k, v in menu.children.items():
            self._write_code_to_tui_file(
                f"self.{k} = self.__class__.{k}"
                f'(service, version, mode, path + ["{v.tui_name}"])\n',
                indent,
            )
        self._write_code_to_tui_file(
            "super().__init__(service, version, mode, path)\n", indent
        )
        indent -= 1

        command_names = [v.name for _, v in menu.children.items() if v.is_command]
        if command_names:
            for command in command_names:
                self._write_code_to_tui_file(f"class {command}(TUIMethod):\n", indent)
                indent += 1
                self._write_code_to_tui_file('"""\n', indent)
                doc_lines = menu.children[command].doc.splitlines()
                for line in doc_lines:
                    line = line.strip()
                    if line:
                        self._write_code_to_tui_file(f"{line}\n", indent)
                self._write_code_to_tui_file('"""\n', indent)
                indent -= 1
        for k, v in menu.children.items():
            if v.is_command:
                api_tree[k] = "Command"
                pass
            else:
                api_tree[k] = self._write_menu_to_tui_file(v, indent)
        return api_tree

    def generate(self) -> None:
        """Generate TUI API classes."""
        api_tree = {}
        Path(self._tui_file).parent.mkdir(exist_ok=True)
        if self._verbose:
            print(f"{str(self._tui_file)}")
        with open(self._tui_file, "w", encoding="utf8") as self.__writer:
            if FluentVersion(self._version) == FluentVersion.v222:
                with open(
                    (
                        Path(__file__)
                        / ".."
                        / "data"
                        / f"static_info_{self._version}_{self._mode}.pickle"
                    ).resolve(),
                    "rb",
                ) as f:
                    self._main_menu = _RenameModuleUnpickler(f).load()
            else:
                info = self._static_infos[
                    (
                        StaticInfoType.TUI_MESHING
                        if self._mode == "meshing"
                        else StaticInfoType.TUI_SOLVER
                    )
                ]
                self._populate_menu(self._main_menu, info)
            self._write_code_to_tui_file(
                f'"""Fluent {self._mode.title().lower()} TUI commands"""\n'
            )
            self._main_menu.doc = f"Fluent {self._mode} main menu."
            self._write_code_to_tui_file(
                "#\n"
                "# This is an auto-generated file.  DO NOT EDIT!\n"
                "#\n"
                "# pylint: disable=line-too-long\n\n"
                "from ansys.fluent.core.services.datamodel_tui "
                "import PyMenu, TUIMenu, TUIMethod\n\n\n"
            )
            self._main_menu.name = "main_menu"
            api_tree["tui"] = self._write_menu_to_tui_file(self._main_menu)
        return api_tree


def generate(version, static_infos: dict, verbose: bool = False):
    """Generate TUI API classes."""
    api_tree = {}
    gt_222 = FluentVersion(version) > FluentVersion.v222
    if gt_222:
        if (
            StaticInfoType.TUI_MESHING not in static_infos
            and StaticInfoType.TUI_SOLVER not in static_infos
        ):
            return api_tree
        _copy_tui_help_xml_file(version)
    _populate_xml_helpstrings()
    if not gt_222 or StaticInfoType.TUI_MESHING in static_infos:
        api_tree["<meshing_session>"] = TUIGenerator(
            "meshing", version, static_infos, verbose
        ).generate()
    if not gt_222 or StaticInfoType.TUI_SOLVER in static_infos:
        api_tree["<solver_session>"] = TUIGenerator(
            "solver", version, static_infos, verbose
        ).generate()
    if os.getenv("PYFLUENT_HIDE_LOG_SECRETS") != "1":
        logger.info(
            "XML help is available but not picked for the following %i paths: ",
            len(_XML_HELPSTRINGS),
        )
        for k in _XML_HELPSTRINGS:
            logger.info(k)
    return api_tree


if __name__ == "__main__":
    solver = launch_fluent()
    meshing = launch_fluent(mode=FluentMode.MESHING)
    version = get_version_for_file_name(session=solver)
    static_infos = {}
    if FluentVersion(version) > FluentVersion.v222:
        static_infos[StaticInfoType.TUI_SOLVER] = (
            solver._datamodel_service_tui.get_static_info("")
        )
        static_infos[StaticInfoType.TUI_MESHING] = (
            meshing._datamodel_service_tui.get_static_info("")
        )
    parser = argparse.ArgumentParser(
        description="A script to write Fluent API files with an optional verbose output."
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Show paths of written Fluent API files.",
    )
    args = parser.parse_args()
    generate(version, static_infos, args.verbose)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing class encapsulating Fluent connection.

Expose icing capabilities.
"""

import importlib
from typing import Any, Dict

from ansys.fluent.core.fluent_connection import FluentConnection
from ansys.fluent.core.services import SchemeEval
from ansys.fluent.core.session_solver import Solver


class SolverIcing(Solver):
    """Encapsulates a Fluent server for Icing session connection.

    SolverIcing(Session) holds the top-level objects for solver TUI, settings and icing
    datamodel objects calls.
    """

    def __init__(
        self,
        fluent_connection: FluentConnection,
        scheme_eval: SchemeEval,
        file_transfer_service: Any | None = None,
        start_transcript: bool = True,
        launcher_args: Dict[str, Any] | None = None,
    ):
        """SolverIcing session.

        Parameters
        ----------
        fluent_connection (:ref:`ref_fluent_connection`):
            Encapsulates a Fluent connection.
        scheme_eval: SchemeEval
            Instance of ``SchemeEval`` to execute Fluent's scheme code on.
        file_transfer_service : Optional
            Service for uploading and downloading files.
        start_transcript : bool, optional
            Whether to start the Fluent transcript in the client.
            The default is ``True``, in which case the Fluent
            transcript can be subsequently started and stopped
            using method calls on the ``Session`` object.
        """
        super(SolverIcing, self).__init__(
            fluent_connection=fluent_connection,
            scheme_eval=scheme_eval,
            file_transfer_service=file_transfer_service,
            start_transcript=start_transcript,
            launcher_args=launcher_args,
        )
        self._flserver_root = None
        self._fluent_version = None
        self._fluent_connection = fluent_connection

    @property
    def _flserver(self):
        """Root datamodel object."""
        if self._flserver_root is None:
            se = self._datamodel_service_se
            dm_module = importlib.import_module(
                f"ansys.fluent.core.datamodel_{self._version}.flicing"
            )
            self._flserver_root = dm_module.Root(se, "flserver", [])
        return self._flserver_root

    @property
    def icing(self):
        """Instance of icing (Case.App) -> root datamodel object."""
        return self._flserver.Case.App

    def __dir__(self):
        return super(SolverIcing, self).__dir__()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from ansys.fluent.core.utils.fluent_version import FluentVersion


@pytest.mark.fluent_version("latest")
def test_creatable(mixing_elbow_case_data_session) -> None:
    solver = mixing_elbow_case_data_session
    fluent_version = solver.get_fluent_version()
    has_not = (
        solver.setup.boundary_conditions.velocity_inlet,
        solver.setup.cell_zone_conditions.fluid,
    )
    has = (
        solver.results.graphics.contour,
        solver.results.graphics.vector,
    )

    for obj in has_not:
        # creatability condition is dynamic since 25.1
        if fluent_version >= FluentVersion.v251:
            assert not getattr(obj, "create").is_active()
        else:
            assert not hasattr(obj, "create")
            assert "create" not in dir(obj)

    for obj in has:
        assert hasattr(obj, "create")
        assert "create" in dir(obj)
        if fluent_version >= FluentVersion.v251:
            assert getattr(obj, "create").is_active()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module to get Fluent version."""

from enum import Enum
from functools import total_ordering
import os

import ansys.fluent.core as pyfluent


class AnsysVersionNotFound(RuntimeError):
    """Raised when Ansys version is not found."""

    pass


class ComparisonError(RuntimeError):
    """Raised when a comparison can't be completed."""

    def __init__(self):
        """Initialize ComparisonError."""
        super().__init__(
            "Comparison operations are only supported between two members of 'FluentVersion'."
        )


def get_version(session=None):
    """Get Fluent version."""
    if session is None:
        session = pyfluent.launch_fluent()

    return session.get_fluent_version().value


def get_version_for_file_name(version: str | None = None, session=None):
    """Get Fluent version for file name."""
    if version is None:
        version = get_version(session)

    return "".join(version.split(".")[0:2])


@total_ordering
class FluentVersion(Enum):
    """An enumeration over supported Fluent versions.

    Examples
    --------
    FluentVersion("23.2.0") == FluentVersion.v232

    FluentVersion.v232.number == 232

    FluentVersion.v232.awp_var == 'AWP_ROOT232'
    """

    v252 = "25.2.0"
    v251 = "25.1.0"
    v242 = "24.2.0"
    v241 = "24.1.0"
    v232 = "23.2.0"
    v231 = "23.1.0"
    v222 = "22.2.0"

    @classmethod
    def _missing_(cls, version):
        if isinstance(version, (int, float, str)):
            version = str(version)
            if len(version) == 3:
                version = version[:2] + "." + version[2:]
            version += ".0"
            for member in cls:
                if version == member.value:
                    return member
        raise AnsysVersionNotFound(
            f"The specified version '{version[:-2]}' is not supported."
            + " Supported versions are: "
            + ", ".join([member.value for member in cls][::-1])
        )

    @classmethod
    def get_latest_installed(cls):
        """Return the version member corresponding to the most recent, available ANSYS
        installation.

        Returns
        -------
        FluentVersion
            FluentVersion member corresponding to the newest Fluent version.

        Raises
        ------
        AnsysVersionNotFound
            If an Ansys version cannot be found.
        """
        for member in cls:
            if member.awp_var in os.environ:
                return member

        raise AnsysVersionNotFound(
            "Verify the value of the 'AWP_ROOT' environment variable."
        )

    @classmethod
    def current_release(cls):
        """Return the version member of the current release.

        Returns
        -------
        FluentVersion
            FluentVersion member corresponding to the latest release.
        """
        return cls(pyfluent.FLUENT_RELEASE_VERSION)

    @classmethod
    def current_dev(cls):
        """Return the version member of the current development version.

        Returns
        -------
        FluentVersion
            FluentVersion member corresponding to the latest development version.
        """
        return cls(pyfluent.FLUENT_DEV_VERSION)

    @property
    def awp_var(self):
        """Get the Fluent version in AWP environment variable format."""
        return f"AWP_ROOT{self.number}"

    @property
    def number(self):
        """Get the Fluent version as a plain integer."""
        return int(self.value.replace(".", "")[:-1])

    @property
    def docker_image_tag(self):
        """Get the Fluent version as a Docker image tag."""
        return f"v{self.value}"

    def __lt__(self, other):
        if isinstance(other, FluentVersion):
            return self.value < other.value
        raise ComparisonError()

    def __repr__(self) -> str:
        """Return a string representation for the Fluent version."""
        return self.value

    def __str__(self) -> str:
        """String output for the Fluent version."""
        return (
            f"Fluent version 20{self.value.split('.')[0]} R{self.value.split('.')[1]}"
        )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Client side implementation of the stand-alone datamodel server."""

from pathlib import Path

import grpc

from ansys.fluent.core.services.datamodel_se import (
    DatamodelService,
    PySimpleMenuGeneric,
)
from tests.run_stateengine_server import kill_server, run_server


def run_datamodel_server(batch_file_name: str | Path, rules):
    """Run the datamodel server."""
    run_command = str(batch_file_name) + " " + rules
    run_server(run_command)
    _channel = grpc.insecure_channel("localhost:50055")
    _data_model_service = DatamodelService(channel=_channel, metadata=[])

    return PySimpleMenuGeneric(_data_model_service, rules)


def close_server():
    """Close the server."""
    grpc.insecure_channel("localhost:50055").close()
    try:
        kill_server()
        print("Server closed successfully.")
    except OSError:
        print(
            "WARNING: Attempt to kill server failed!!! \nPlease kill the server manually."
        )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Unit tests for flobject module."""

from collections.abc import MutableMapping
import io
import weakref

import pytest
from test_utils import count_key_recursive

from ansys.fluent.core.examples import download_file
from ansys.fluent.core.solver import flobject
from ansys.fluent.core.solver.flobject import (
    InactiveObjectError,
    _gethash,
    find_children,
)
from ansys.fluent.core.utils.fluent_version import FluentVersion
import ansys.units


class Setting:
    """Base class for setting objects."""

    def __init__(self, parent):
        self.parent = None if parent is None else weakref.proxy(parent)

    def get_attr(self, attr):
        attrs = self.get_attrs([attr])
        if attrs.get("active?"):
            return attrs[attr]
        else:
            raise RuntimeError("Object is not active")

    def get_attrs(self, attrs):
        active = self.attrs.get("active?", lambda self: True)(self)
        if active:
            return {attr: self.attrs[attr](self) for attr in attrs}
        else:
            return {"active?": False}

    attrs = {
        "active?": lambda self: True,
        "webui-release-active?": lambda self: True,
        "deprecated-version": lambda self: None,
    }


class PrimitiveSetting(Setting):
    """Primitive setting objects."""

    value = None

    def get_state(self):
        return self.value

    def set_state(self, value):
        self.value = value

    @classmethod
    def get_static_info(cls):
        ret = {"type": cls.objtype}
        if cls.__doc__:
            ret["help"] = cls.__doc__
        return ret


class Bool(PrimitiveSetting):
    objtype = "boolean"


class Int(PrimitiveSetting):
    objtype = "integer"


class Real(PrimitiveSetting):
    objtype = "real"


class String(PrimitiveSetting):
    objtype = "string/symbol"


class BoolList(PrimitiveSetting):
    objtype = "boolean-list"


class IntList(PrimitiveSetting):
    objtype = "integer-list"


class RealList(PrimitiveSetting):
    objtype = "real-list"


class StringList(PrimitiveSetting):
    objtype = "string-list"


class Group(Setting):
    """Group objects."""

    objtype = "group"
    children = {}
    commands = {}

    def __init__(self, parent):
        super().__init__(parent)
        self.objs = {c: v(self) for c, v in self.children.items()}

    def get_state(self):
        ret = {}
        for c in self.children:
            cobj = self.objs[c]
            if cobj.get_attr("active?"):
                ret[c] = cobj.get_state()
        return ret

    def set_state(self, value):
        for c in self.children:
            v = value.get(c)
            if v is not None:
                self.objs[c].set_state(v)

    def get_child(self, c):
        return self.objs[c]

    def get_command(self, c):
        return self.commands[c](self)

    @classmethod
    def get_static_info(cls):
        ret = {"type": cls.objtype}
        if cls.__doc__:
            ret["help"] = cls.__doc__
        if cls.children:
            ret["children"] = {c: v.get_static_info() for c, v in cls.children.items()}
        if cls.commands:
            ret["commands"] = {c: v.get_static_info() for c, v in cls.commands.items()}
        return ret


class NamedObject(Setting, MutableMapping):
    """NamedObject class."""

    objtype = "named-object"
    commands = {}
    # To be overridden by child classes
    # child_object_type = None

    def __init__(self, parent):
        super().__init__(parent)
        self._objs = {}

    def __getitem__(self, name):
        return self._objs[name].get_state()

    def __setitem__(self, name, value):
        if name not in self._objs:
            self._objs[name] = self.child_object_type(self)
        return self._objs[name].set_state(value)

    def __delitem__(self, name):
        del self._objs[name]

    def __iter__(self):
        return iter(self._objs)

    def __len__(self):
        return len(self._objs)

    def get_child(self, c):
        return self._objs[c]

    def rename(self, new, old):
        self._objs = {(new if k == old else k): v for k, v in self._objs.items()}

    def get_object_names(self):
        return list(self._objs.keys())

    def get_command(self, c):
        return self.commands[c](self)

    def get_state(self):
        return {c: v.get_state() for c, v in self._objs.items()}

    def set_state(self, state):
        for k, v in state.items():
            self[k] = v

    @classmethod
    def get_static_info(cls):
        ret = {"type": cls.objtype}
        if cls.__doc__:
            ret["help"] = cls.__doc__
        ret["object-type"] = cls.child_object_type.get_static_info()
        if cls.commands:
            ret["commands"] = {c: v.get_static_info() for c, v in cls.commands.items()}
        try:
            if cls.user_creatable:
                ret["user_creatable"] = cls.user_creatable
        except AttributeError:
            ret["user_creatable"] = True
        return ret


class ListObject(Setting):
    """ListObject class."""

    objtype = "list-object"
    commands = {}
    # To be overridden by child classes
    # child_object_type = None

    def __init__(self, parent):
        super().__init__(parent)
        self._objs = []

    def __getitem__(self, index):
        return self._objs[index].get_state()

    def __setitem__(self, index, value):
        return self._objs[index].set_state(value)

    def __iter__(self):
        return iter(self._objs)

    def __len__(self):
        return len(self._objs)

    def size(self):
        return len(self._objs)

    def resize(self, new_size):
        if new_size > len(self._objs):
            for _ in range(len(self._objs), new_size):
                self._objs.append(self.child_object_type(self))
        elif new_size < len(self._objs):
            self._objs = self._objs[:new_size]

    def get_child(self, c):
        return self._objs[int(c)]

    def get_command(self, c):
        return self.commands[c](self)

    def get_state(self):
        return [x.get_state() for x in self._objs]

    def set_state(self, value):
        self.resize(len(value))
        for i, v in enumerate(value):
            self[i] = v

    @classmethod
    def get_static_info(cls):
        ret = {"type": cls.objtype}
        if cls.__doc__:
            ret["help"] = cls.__doc__
        ret["object-type"] = cls.child_object_type.get_static_info()
        if cls.commands:
            ret["commands"] = {c: v.get_static_info() for c, v in cls.commands.items()}
        return ret


class Command(Setting):
    """Command class."""

    objtype = "command"
    # To be overridden by child classes
    # arguments = None
    # cb = None

    def __init__(self, parent):
        self.attrs = super().attrs.copy()
        self.attrs["arguments-aliases"] = lambda self: {}
        super().__init__(parent)

    def __call__(self, **kwds):
        args = []
        for k, v in self.arguments.items():
            a = kwds.get(k, v(self).get_state())
            args.append(a)
        return self.cb(*args)

    @classmethod
    def get_static_info(cls):
        ret = {"type": cls.objtype}
        if cls.__doc__:
            ret["help"] = cls.__doc__
        if cls.arguments:
            ret["arguments"] = {
                c: v.get_static_info() for c, v in cls.arguments.items()
            }
        return ret


class Root(Group):
    """Root class."""

    class G1(Group):
        class S1(String):
            attrs = {
                "active?": lambda self: not self.parent.objs["b-3"].get_state(),
                "allowed-values": lambda self: ["foo", "bar"],
                "webui-release-active?": lambda self: True,
                "deprecated-version": lambda self: None,
            }

        children = {
            "r-1": Real,
            "i-2": Int,
            "b-3": Bool,
            "s-4": S1,
        }

    class N1(NamedObject):
        class NC(Group):
            children = {
                "rl-1": RealList,
                "sl-1": StringList,
            }

        child_object_type = NC

    class L1(ListObject):
        class LC(Group):
            children = {
                "il-1": IntList,
                "bl-1": BoolList,
            }

        child_object_type = LC

    class Command1(Command):
        """Command1 class."""

        class A1(Real):
            value = 2.3

        class A2(Bool):
            value = True

        arguments = {
            "a-1": A1,
            "a-2": A2,
        }

        def cb(self, a1, a2):
            if a2 is True:
                self.parent.objs["g-1"].objs["r-1"].value += a1
            else:
                self.parent.objs["g-1"].objs["r-1"].value -= a1

    children = {
        "g-1": G1,
        "n-1": N1,
        "l-1": L1,
    }

    commands = {
        "c-1": Command1,
    }


class Proxy:
    """Proxy class."""

    root = Root

    def __init__(self):
        self.r = self.root(None)

    def get_obj(self, path):
        if not path:
            return self.r
        obj = self.r
        for c in path.split("/"):
            try:
                obj = obj.get_child(c)
            except KeyError:
                obj = obj.get_command(c)
        return obj

    def get_var(self, path):
        return self.get_obj(path).get_state()

    def set_var(self, path, value):
        return self.get_obj(path).set_state(value)

    def rename(self, path, new, old):
        return self.get_obj(path).rename(new, old)

    def create(self, path, name):
        self.get_obj(path)[name] = {}

    def delete(self, path, name):
        del self.get_obj(path)[name]

    def resize_list_object(self, path, size):
        return self.get_obj(path).resize(size)

    def get_list_size(self, path):
        return self.get_obj(path).size()

    def get_object_names(self, path):
        return self.get_obj(path).get_object_names()

    def execute_cmd(self, path, command, **kwds):
        return self.get_obj(path).get_command(command)(**kwds)

    def get_attrs(self, path, attrs, recursive=False):
        return self.get_obj(path).get_attrs(attrs)

    @classmethod
    def get_static_info(cls):
        return cls.root.get_static_info()

    def is_interactive_mode(self):
        return False

    def has_wildcard(self, name: str) -> bool:
        return False


def test_primitives():
    r = flobject.get_root(Proxy())
    r.g_1.r_1 = 3.2
    assert r.g_1.r_1() == 3.2
    r.g_1.i_2 = -3
    assert r.g_1.i_2() == -3
    r.g_1.b_3 = True
    assert r.g_1.b_3() is True
    r.g_1.b_3 = False
    assert r.g_1.b_3() is False
    r.g_1.s_4 = "foo"
    assert r.g_1.s_4() == "foo"


def test_group():
    r = flobject.get_root(Proxy())
    r.g_1 = {"r_1": 3.2, "i_2": -3, "b_3": False, "s_4": "foo"}
    assert r.g_1() == {"r_1": 3.2, "i_2": -3, "b_3": False, "s_4": "foo"}
    r.g_1 = {"s_4": "bar"}
    assert r.g_1() == {"r_1": 3.2, "i_2": -3, "b_3": False, "s_4": "bar"}
    r.g_1.i_2 = 4
    assert r.g_1() == {"r_1": 3.2, "i_2": 4, "b_3": False, "s_4": "bar"}


def test_settings_input_set_state():
    r = flobject.get_root(Proxy())
    r.g_1 = {"r_1": 3.2, "i_2": -3, "b_3": False, "s_4": "foo"}
    r.g_1.set_state(r_1=3.2, i_2=-3, b_3=False, s_4="foo")
    assert r.g_1() == {"r_1": 3.2, "i_2": -3, "b_3": False, "s_4": "foo"}
    r.g_1.set_state(s_4="bar")
    assert r.g_1() == {"r_1": 3.2, "i_2": -3, "b_3": False, "s_4": "bar"}
    r.g_1.set_state(i_2=4)
    assert r.g_1() == {"r_1": 3.2, "i_2": 4, "b_3": False, "s_4": "bar"}


def test_settings_input():
    r = flobject.get_root(Proxy())
    r.g_1 = {"r_1": 3.2, "i_2": -3, "b_3": False, "s_4": "foo"}
    r.g_1(r_1=3.2, i_2=-3, b_3=False, s_4="foo")
    assert r.g_1() == {"r_1": 3.2, "i_2": -3, "b_3": False, "s_4": "foo"}
    r.g_1(s_4="bar")
    assert r.g_1() == {"r_1": 3.2, "i_2": -3, "b_3": False, "s_4": "bar"}
    r.g_1(i_2=4)
    assert r.g_1() == {"r_1": 3.2, "i_2": 4, "b_3": False, "s_4": "bar"}


def test_named_object():
    r = flobject.get_root(Proxy())
    assert r.n_1.get_object_names() == []
    r.n_1["n1"] = {}
    r.n_1["n2"] = {}
    assert r.n_1.get_object_names() == ["n1", "n2"]
    r.n_1.create("n4")
    assert r.n_1.get_object_names() == ["n1", "n2", "n4"]
    del r.n_1["n1"]
    assert r.n_1.get_object_names() == ["n2", "n4"]
    r.n_1["n1"] = {"rl_1": [1.2, 3.4], "sl_1": ["foo", "bar"]}
    assert r.n_1["n1"]() == {"rl_1": [1.2, 3.4], "sl_1": ["foo", "bar"]}
    r.n_1 = {"n5": {"rl_1": [4.3, 2.1], "sl_1": ["oof", "rab"]}}
    assert r.n_1.get_object_names() == ["n2", "n4", "n1", "n5"]
    assert r.n_1["n5"]() == {"rl_1": [4.3, 2.1], "sl_1": ["oof", "rab"]}


def test_list_object():
    r = flobject.get_root(Proxy())
    assert r.l_1.get_size() == 0
    r.l_1 = [
        {"il_1": None, "bl_1": None},
        {"il_1": None, "bl_1": None},
    ]
    r.l_1[1].il_1 = [1, 2]
    assert r.l_1() == [
        {"il_1": None, "bl_1": None},
        {"il_1": [1, 2], "bl_1": None},
    ]
    r.l_1 = [{"il_1": [3], "bl_1": [True, False]}]
    assert r.l_1() == [{"il_1": [3], "bl_1": [True, False]}]


def test_command():
    r = flobject.get_root(Proxy())
    r.g_1.r_1 = 2.4
    r.c_1()
    assert r.g_1.r_1() == 2.4 + 2.3
    r.c_1(a_2=False)
    assert r.g_1.r_1() == 2.4 + 2.3 - 2.3
    r.c_1(a_1=3.2, a_2=True)
    assert r.g_1.r_1() == 2.4 + 2.3 - 2.3 + 3.2
    r.c_1(a_1=4.5, a_2=False)
    assert r.g_1.r_1() == 2.4 + 2.3 - 2.3 + 3.2 - 4.5


def test_attrs():
    r = flobject.get_root(Proxy())
    r._setattr("_version", "251")
    assert r.g_1.s_4.get_attr("active?")
    assert r.g_1.s_4.get_attr("allowed-values") == ["foo", "bar"]
    r.g_1.b_3 = True
    assert not r.g_1.s_4.get_attr("active?")
    with pytest.raises(InactiveObjectError):
        r.g_1.s_4.get_attr("allowed-values")


# The following test is commented out as codegen module is not packaged in the
# install
def _disabled_test_settings_gen():
    info = Proxy().get_static_info()
    cls, _ = flobject.get_cls("", info)
    f = io.StringIO()
    ansys.fluent.core.codegen.settingsgen.write_settings_classes(f, cls, info)
    assert (
        f.getvalue()
        == '''###
### THIS FILE IS AUTOGENERATED! DO NOT MODIFY!

###
from ansys.fluent.solver.flobject import *

SHASH = "0392eb93ff1d5f9dd50ef9fddc74581ad3e8f74c34c569b6a1a6f1b57753d5ad"

class root(Group):
    """
    Root class
    """
    fluent_name = ""
    child_names = \\
        ['g_1', 'n_1', 'l_1']

    class g_1(Group):
        """
        'g_1' child of 'root' object
        """
        fluent_name = "g-1"
        child_names = \\
            ['r_1', 'i_2', 'b_3', 's_4']

        class r_1(Real):
            """
            'r_1' child of 'g_1' object
            """
            fluent_name = "r-1"

        class i_2(Integer):
            """
            'i_2' child of 'g_1' object
            """
            fluent_name = "i-2"

        class b_3(Boolean):
            """
            'b_3' child of 'g_1' object
            """
            fluent_name = "b-3"

        class s_4(String):
            """
            's_4' child of 'g_1' object
            """
            fluent_name = "s-4"

    class n_1(NamedObject):
        """
        'n_1' child of 'root' object
        """
        fluent_name = "n-1"

        class child_object_type(Group):
            """
            'child_object_type' child of 'n_1' object
            """
            fluent_name = "child-object-type"
            child_names = \\
                ['rl_1', 'sl_1']

            class rl_1(RealList):
                """
                'rl_1' child of 'child_object_type' object
                """
                fluent_name = "rl-1"

            class sl_1(StringList):
                """
                'sl_1' child of 'child_object_type' object
                """
                fluent_name = "sl-1"

    class l_1(ListObject):
        """
        'l_1' child of 'root' object
        """
        fluent_name = "l-1"

        class child_object_type(Group):
            """
            'child_object_type' child of 'l_1' object
            """
            fluent_name = "child-object-type"
            child_names = \\
                ['il_1', 'bl_1']

            class il_1(IntegerList):
                """
                'il_1' child of 'child_object_type' object
                """
                fluent_name = "il-1"

            class bl_1(BooleanList):
                """
                'bl_1' child of 'child_object_type' object
                """
                fluent_name = "bl-1"
    command_names = \\
        ['c_1']

    class c_1(Command):
        """
        Command1 class

        Parameters
        ----------
            a_1 : real
                'a_1' child of 'c_1' object
            a_2 : bool
                'a_2' child of 'c_1' object

        """
        fluent_name = "c-1"
        argument_names = \\
            ['a_1', 'a_2']

        class a_1(Real):
            """
            'a_1' child of 'c_1' object
            """
            fluent_name = "a-1"

        class a_2(Boolean):
            """
            'a_2' child of 'c_1' object
            """
            fluent_name = "a-2"
'''
    )  # noqa: W293


@pytest.mark.fluent_version("latest")
def test_accessor_methods_on_settings_object(static_mixer_settings_session):
    solver = static_mixer_settings_session

    existing = solver.file.read.file_type.get_attr("allowed-values")
    modified = solver.file.read.file_type.allowed_values()
    assert existing == modified

    existing = solver.file.read.file_type.get_attr("read-only?", bool)
    modified = solver.file.read.file_type.is_read_only()

    assert existing == modified

    velocity_inlet = solver.setup.boundary_conditions.velocity_inlet
    existing = velocity_inlet.get_attr("user-creatable?", bool)
    modified = velocity_inlet.user_creatable()
    assert existing == modified

    if solver.get_fluent_version() < FluentVersion.v242:
        turbulent_viscosity_ratio = velocity_inlet[
            "inlet1"
        ].turbulence.turbulent_viscosity_ratio_real

        path = '<session>.setup.boundary_conditions.velocity_inlet["inlet1"].turbulence.turbulent_viscosity_ratio_real'
        name = "turbulent_viscosity_ratio_real"

    else:
        turbulent_viscosity_ratio = velocity_inlet[
            "inlet1"
        ].turbulence.turbulent_viscosity_ratio

        if solver.get_fluent_version() >= FluentVersion.v251:
            path = '<session>.settings.setup.boundary_conditions.velocity_inlet["inlet1"].turbulence.turbulent_viscosity_ratio'
        else:
            path = '<session>.setup.boundary_conditions.velocity_inlet["inlet1"].turbulence.turbulent_viscosity_ratio'
        name = "turbulent_viscosity_ratio"

    assert turbulent_viscosity_ratio.python_path == path
    assert turbulent_viscosity_ratio.python_name == name

    assert turbulent_viscosity_ratio.default_value() == 10
    assert turbulent_viscosity_ratio.get_attr("min") == 0

    assert turbulent_viscosity_ratio.get_attr("max") is False
    assert turbulent_viscosity_ratio.max() is None

    default_attrs = solver.setup.boundary_conditions.velocity_inlet["inlet1"].get_attrs(
        ["default"], recursive=True
    )
    assert count_key_recursive(default_attrs, "default") > 5

    mesh = solver.results.graphics.mesh.create("mesh-1")
    if solver.get_fluent_version() < FluentVersion.v242:
        assert mesh.name.is_read_only()
    else:
        assert not mesh.name.is_read_only()

    assert solver.results.graphics.mesh.get_object_names() == ["mesh-1"]

    solver.results.graphics.mesh["mesh-1"].rename("mesh_new")
    assert solver.results.graphics.mesh.get_object_names() == ["mesh_new"]

    solver.results.graphics.mesh.rename(new="mesh_242", old="mesh_new")
    assert solver.results.graphics.mesh.get_object_names() == ["mesh_242"]


@pytest.mark.fluent_version("latest")
def test_accessor_methods_on_settings_object_types(static_mixer_settings_session):
    solver = static_mixer_settings_session

    assert solver.setup.general.solver.type.allowed_values() == [
        "pressure-based",
        "density-based-implicit",
        "density-based-explicit",
    ]
    accuracy_control = (
        solver.setup.models.discrete_phase.numerics.tracking.accuracy_control
    )
    if solver.get_fluent_version() < FluentVersion.v241:
        max_refinements = accuracy_control.max_number_of_refinements
    else:
        max_refinements = accuracy_control.max_num_refinements

    assert max_refinements.min() == 0
    assert max_refinements.max() == 1000000
    assert max_refinements.get_attr("max") == 1000000


@pytest.mark.fluent_version("==24.1")
@pytest.mark.codegen_required
def test_find_children_from_settings_root(static_mixer_settings_session):
    setup_cls = static_mixer_settings_session.setup.__class__
    assert len(find_children(setup_cls())) >= 10000
    assert len(find_children(setup_cls(), "gen*")) >= 9
    assert set(find_children(setup_cls(), "general*")) >= {
        "general",
        "models/discrete_phase/general_settings",
        "models/virtual_blade_model/rotor/general",
    }
    assert set(find_children(setup_cls(), "general")) >= {
        "general",
        "models/virtual_blade_model/rotor/general",
    }
    assert any(
        path
        for path in find_children(setup_cls(), "*gen")
        if path.endswith("p_backflow_spec_gen")
    )


@pytest.mark.fluent_version("latest")
def test_find_children_from_fluent_solver_session(static_mixer_settings_session):
    setup_children = find_children(static_mixer_settings_session.setup)
    load_mixer = static_mixer_settings_session.setup
    assert len(setup_children) >= 18514

    viscous = load_mixer.models.viscous
    assert len(find_children(viscous, "prod*")) > 0

    assert any(
        path
        for path in find_children(
            load_mixer.boundary_conditions.pressure_outlet, "*_dir_*"
        )
        if path.endswith("geom_dir_spec")
    )

    if static_mixer_settings_session.get_fluent_version() < FluentVersion.v242:
        assert set(
            find_children(
                load_mixer.materials.fluid["air"].density.piecewise_polynomial
            )
        ) >= {
            "minimum",
            "maximum",
            "coefficients",
        }
    else:
        assert set(
            find_children(
                load_mixer.materials.fluid["air"].density.piecewise_polynomial
            )
        ) >= {
            "range/minimum",
            "range/maximum",
            "range/coefficients",
        }


@pytest.mark.fluent_version(">=24.1")
def test_settings_wild_card_access(new_solver_session) -> None:
    solver = new_solver_session

    case_path = download_file("elbow_source_terms.cas.h5", "pyfluent/mixing_elbow")
    solver.file.read_case(file_name=case_path)

    solver.solution.initialization.hybrid_initialize()

    if solver.get_fluent_version() >= FluentVersion.v251:
        assert (
            solver.setup.boundary_conditions.velocity_inlet[
                "*1"
            ].momentum.velocity_magnitude.value()["inlet1"]["momentum"][
                "velocity_magnitude"
            ][
                "value"
            ]
            == solver.setup.boundary_conditions.velocity_inlet[
                "inlet1"
            ].momentum.velocity.value()
        )
    else:
        assert (
            solver.setup.boundary_conditions.velocity_inlet[
                "*1"
            ].momentum.velocity.value()["inlet1"]["momentum"]["velocity"]["value"]
            == solver.setup.boundary_conditions.velocity_inlet[
                "inlet1"
            ].momentum.velocity.value()
        )

    assert solver.setup.boundary_conditions.wall["*"]()

    with pytest.raises(AttributeError) as msg:
        solver.setup.boundary_conditions.velocity_inlet["*1"].inlet1()
    assert msg.value.args[0] == "'velocity_inlet' has no attribute 'inlet1'.\n"

    with pytest.raises(KeyError) as msg:
        solver.setup.boundary_conditions.velocity_inlet["inlet-1"]
    assert (
        msg.value.args[0] == "'velocity_inlet' has no attribute 'inlet-1'.\n"
        "The most similar names are: inlet1, inlet2."
    )


@pytest.mark.fluent_version(">=25.1")
def test_settings_matching_names(new_solver_session) -> None:
    solver = new_solver_session

    case_path = download_file("elbow_source_terms.cas.h5", "pyfluent/mixing_elbow")
    solver.file.read_case(file_name=case_path)

    solver.solution.initialization.hybrid_initialize()

    with pytest.raises(AttributeError) as msg:
        solver.setup.mod

    assert msg.value.args[0].startswith("'setup' object has no attribute 'mod'.")

    with pytest.raises(ValueError) as msg:
        solver.setup.models.viscous.model = "k_epsilon"

    assert (
        msg.value.args[0] == "'model' has no attribute 'k_epsilon'.\n"
        "The most similar names are: k-epsilon."
    )


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.2")
def test_settings_api_names_exception(new_solver_session):
    solver = new_solver_session

    case_path = download_file("mixing_elbow.msh.h5", "pyfluent/mixing_elbow")
    solver.file.read_case(file_name=case_path)

    with pytest.raises(RuntimeError):
        solver.setup.boundary_conditions["cold-inlet"].name = "hot-inlet"


@pytest.mark.fluent_version(">=24.2")
def test_accessor_methods_on_settings_objects(new_solver_session):
    solver = new_solver_session
    root = solver.settings

    nodes = {}
    expected_type_list = [
        "Boolean",
        "String",
        "Real",
        "Integer",
        "RealList",
        "ListObject",
    ]
    type_list = expected_type_list.copy()

    get_child_nodes(root, nodes, type_list)

    for type_data in expected_type_list:
        if type_data == "Boolean":
            assert {
                "is_active",
                "is_read_only",
                "default_value",
                "get_state",
                "set_state",
            }.issubset(set(dir(nodes[type_data])))
            assert nodes[type_data].is_read_only() in [True, False]
            assert nodes[type_data].is_active() in [True, False]

        elif type_data in ["Integer", "Real", "IntegerList", "RealList"]:
            assert {
                "is_active",
                "is_read_only",
                "default_value",
                "get_state",
                "set_state",
                "min",
                "max",
            }.issubset(set(dir(nodes[type_data])))
            assert not {"allowed_values"}.issubset(set(dir(nodes[type_data])))
            assert nodes[type_data].is_read_only() in [True, False]
            assert nodes[type_data].is_active() in [True, False]

        elif type_data in ["String", "StringList", "Filename"]:
            assert {
                "is_active",
                "is_read_only",
                "default_value",
                "get_state",
                "set_state",
                "allowed_values",
            }.issubset(set(dir(nodes[type_data])))
            assert not {"min", "max"}.issubset(set(dir(nodes[type_data])))
            assert nodes[type_data].is_read_only() in [True, False]
            assert nodes[type_data].is_active() in [True, False]

        elif type_data == "ListObject":
            assert {"is_active", "is_read_only", "get_state", "set_state"}.issubset(
                set(dir(nodes[type_data]))
            )
            assert nodes[type_data].is_read_only() in [True, False]
            assert nodes[type_data].is_active() in [True, False]


def get_child_nodes(node, nodes, type_list):
    if node.is_active():
        if isinstance(node, flobject.Group):
            for item in node.child_names:
                get_child_nodes(getattr(node, item), nodes, type_list)
        else:
            node_type = node.__class__.__bases__[0].__name__
            if node_type in type_list:
                type_list.remove(node_type)
                nodes[node_type] = node
                if not type_list:
                    return


@pytest.mark.fluent_version("latest")
def test_strings_with_allowed_values(static_mixer_settings_session):
    solver = static_mixer_settings_session

    with pytest.raises(AttributeError) as e:
        solver.file.auto_save.root_name.allowed_values()
    assert e.value.args[0] == "'root_name' object has no attribute 'allowed_values'"

    string_with_allowed_values = solver.setup.general.solver.type.allowed_values()
    assert string_with_allowed_values == [
        "pressure-based",
        "density-based-implicit",
        "density-based-explicit",
    ]


@pytest.mark.fluent_version(">=24.2")
def test_parent_class_attributes(static_mixer_settings_session):
    solver = static_mixer_settings_session
    assert solver.setup.models.energy.enabled
    with pytest.raises(AttributeError):
        solver.setup.models.energy.__class__.enabled


def _check_vector_units(obj, units):
    assert obj.units() == units
    state_with_units = obj.state_with_units()
    state = obj.get_state()
    assert len(state_with_units) == 2
    assert len(state) == len(state_with_units[0])
    assert all(x == y for x, y in zip(state, state_with_units[0]))
    assert units == state_with_units[1]
    assert obj.as_quantity() == ansys.units.Quantity(obj.get_state(), units)


@pytest.mark.fluent_version(">=24.1")
def test_ansys_units_integration(mixing_elbow_settings_session):
    solver = mixing_elbow_settings_session
    assert isinstance(solver.settings.state_with_units(), dict)
    hot_inlet = solver.setup.boundary_conditions.velocity_inlet["hot-inlet"]
    turbulence = hot_inlet.turbulence
    turbulence.turbulent_specification = "Intensity and Hydraulic Diameter"
    hydraulic_diameter = turbulence.hydraulic_diameter
    hydraulic_diameter.set_state("1 [in]")
    assert hydraulic_diameter() == "1 [in]"
    assert hydraulic_diameter.as_quantity() is None
    assert hydraulic_diameter.state_with_units() == ("1 [in]", "m")
    assert hydraulic_diameter.units() == "m"
    turbulent_intensity = turbulence.turbulent_intensity
    turbulent_intensity.set_state(0.2)
    assert turbulent_intensity() == 0.2
    assert turbulent_intensity.as_quantity() == ansys.units.Quantity(0.2, "")
    turbulent_intensity.set_state(ansys.units.Quantity(0.1, ""))
    assert turbulent_intensity.state_with_units() == (0.1, "")
    hydraulic_diameter.set_state(1)
    assert hydraulic_diameter.as_quantity() == ansys.units.Quantity(1, "m")
    assert hydraulic_diameter.state_with_units() == (1.0, "m")
    assert hydraulic_diameter.units() == "m"
    hydraulic_diameter.set_state(ansys.units.Quantity(1, "in"))
    assert hydraulic_diameter.as_quantity() == ansys.units.Quantity(0.0254, "m")
    assert hydraulic_diameter.state_with_units() == (0.0254, "m")
    assert hydraulic_diameter.units() == "m"
    assert hydraulic_diameter() == 0.0254
    velocity = ansys.units.Quantity(
        12.0, ansys.units.UnitRegistry().ft
    ) / ansys.units.Quantity(3.0, ansys.units.UnitRegistry().s)
    hot_inlet.momentum.velocity.value = velocity
    assert hot_inlet.momentum.velocity.value.as_quantity() == velocity
    velocity = (1.0, "m s^-1")
    hot_inlet.momentum.velocity = velocity
    assert hot_inlet.momentum.velocity.value.state_with_units() == velocity
    velocity = ansys.units.Quantity(12.0, "m s^-1")
    hot_inlet.momentum.velocity = velocity
    assert hot_inlet.momentum.velocity.value() == velocity.value
    assert hot_inlet.momentum.velocity.value.as_quantity() == velocity
    assert hot_inlet.momentum.velocity.state_with_units() == {
        "option": "value",
        "value": (12.0, "m s^-1"),
    }
    # https://github.com/ansys/pyfluent/issues/3738
    # clip_factor = solver.setup.models.viscous.options.production_limiter.clip_factor
    # clip_factor.set_state(1.2)
    # assert clip_factor() == 1.2
    # assert clip_factor.as_quantity() == ansys.units.Quantity(1.2, "")
    # assert clip_factor.state_with_units() == (1.2, "")
    # assert clip_factor.units() == ""
    # clip_factor.set_state(ansys.units.Quantity(1.8, ""))
    # assert clip_factor.as_quantity() == ansys.units.Quantity(1.8, "")
    # assert clip_factor.state_with_units() == (1.8, "")
    # assert clip_factor.units() == ""

    _check_vector_units(
        solver.setup.general.operating_conditions.reference_pressure_location, "m"
    )
    _check_vector_units(
        solver.setup.reference_frames[
            "global"
        ].initial_state.orientation.first_axis.axis_to.vector,
        "",
    )


@pytest.mark.fluent_version(">=24.2")
def test_ansys_units_integration_nested_state(mixing_elbow_settings_session):
    solver = mixing_elbow_settings_session

    hot_inlet = solver.setup.boundary_conditions.velocity_inlet["hot-inlet"]

    assert hot_inlet.state_with_units() == {
        "momentum": {
            "initial_gauge_pressure": {"option": "value", "value": (0, "Pa")},
            "reference_frame": "Absolute",
            "velocity": {"option": "value", "value": (0, "m s^-1")},
            "velocity_specification_method": "Magnitude, Normal to Boundary",
        },
        "name": "hot-inlet",
        "turbulence": {
            "turbulent_intensity": (0.05, ""),
            "turbulent_specification": "Intensity and Viscosity Ratio",
            "turbulent_viscosity_ratio": (10, None),
        },
    } or {
        "momentum": {
            "initial_gauge_pressure": {"option": "value", "value": (0, "Pa")},
            "reference_frame": "Absolute",
            "velocity": {"option": "value", "value": (0, "m s^-1")},
            "velocity_specification_method": "Magnitude, Normal to Boundary",
        },
        "name": "hot-inlet",
        "turbulence": {
            "turbulent_specification": "Intensity and Viscosity Ratio",
            "turbulent_intensity": (0.05, ""),
            "turbulent_viscosity_ratio": (10, None),
        },
    }


@pytest.mark.fluent_version(">=24.2")
def test_bug_1001124_quantity_assignment(mixing_elbow_settings_session):
    speed = ansys.units.Quantity(100, "m s^-1")
    solver = mixing_elbow_settings_session
    solver.setup.boundary_conditions.velocity_inlet[
        "hot-inlet"
    ].momentum.velocity.value = speed.value
    assert (
        solver.setup.boundary_conditions.velocity_inlet[
            "hot-inlet"
        ].momentum.velocity.value()
        == speed.value
    )
    solver.setup.boundary_conditions.velocity_inlet["hot-inlet"].momentum.velocity = (
        speed
    )
    assert (
        solver.setup.boundary_conditions.velocity_inlet[
            "hot-inlet"
        ].momentum.velocity.value()
        == speed.value
    )


def test_assert_type():
    types = [
        bool,
        int,
        flobject.RealType,
        str,
        flobject.BoolListType,
        flobject.IntListType,
        flobject.RealListType,
        flobject.StringListType,
        flobject.RealVectorType,
        flobject.DictStateType,
    ]
    vals = [
        False,
        1,
        1.0,
        "a",
        [False, True],
        [1, 2],
        [1.0, 2.0],
        ["a", "b"],
        (1.0, 2.0, 3.0),
        {"a": 1},
    ]
    subtypes = {
        bool: (int,),
        str: (flobject.RealType,),
        flobject.BoolListType: (flobject.IntListType,),
        flobject.StringListType: (flobject.RealListType,),
    }
    for i_t, tp in enumerate(types):
        for i_v, val in enumerate(vals):
            if i_t == i_v:
                flobject.assert_type(val, tp)
            else:
                subtype = subtypes.get(types[i_v])
                if subtype and types[i_t] in subtype:
                    flobject.assert_type(val, tp)
                else:
                    with pytest.raises(TypeError):
                        flobject.assert_type(val, tp)


def test_static_info_hash_identity(new_solver_session):
    solver = new_solver_session
    hash1 = _gethash(solver._settings_service.get_static_info())
    hash2 = _gethash(solver._settings_service.get_static_info())
    assert hash1 == hash2


@pytest.mark.codegen_required
def test_no_hash_mismatch(new_solver_session, caplog):
    caplog.clear()
    new_solver_session.setup
    assert all(["Mismatch" not in record.message for record in caplog.records])


@pytest.mark.fluent_version(">=24.2")
def test_default_argument_names_for_commands(static_mixer_settings_session):
    solver = static_mixer_settings_session

    if solver.get_fluent_version() >= FluentVersion.v251:
        assert set(solver.results.graphics.contour.command_names) == {
            "create",
            "delete",
            "rename",
            "list",
            "list_properties",
            "make_a_copy",
            "display",
            "add_to_graphics",
            "clear_history",
        }
    else:
        assert set(solver.results.graphics.contour.command_names) == {
            "delete",
            "rename",
            "list",
            "list_properties",
            "make_a_copy",
            "display",
            "copy",
            "add_to_graphics",
            "clear_history",
        }

    assert solver.results.graphics.contour.rename.argument_names == ["new", "old"]
    assert solver.results.graphics.contour.delete.argument_names == ["name_list"]
    # The following is the default behavior when no arguments are associated with the command.
    assert solver.results.graphics.contour.list.argument_names == []



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Reader for Fluent case files.

Example
-------

.. code-block:: python

    >>> from ansys.fluent.core import examples
    >>> from ansys.fluent.core.filereader.case_file import CaseFile

    >>> case_file_name = examples.download_file("Static_Mixer_Parameters.cas.h5", "pyfluent/static_mixer", return_without_path=False)

    >>> reader = CaseFile(case_file_name=case_file_name) # Instantiate a CaseFile class
    >>> input_parameters = reader.input_parameters()     # Get lists of input parameters
    >>> output_parameters = reader.output_parameters()   # Get lists of output parameters
"""

import codecs
from enum import Enum
import gzip
import os
from os.path import dirname
from pathlib import Path
from typing import Dict, List

import defusedxml.ElementTree as ET
import numpy as np

from ansys.fluent.core.solver.error_message import allowed_name_error_message

from . import lispy
from .pre_processor import remove_unsupported_xml_chars

try:
    import h5py
except ModuleNotFoundError as exc:
    raise ModuleNotFoundError(
        "Missing dependencies, use 'pip install ansys-fluent-core[reader]' to install them."
    ) from exc


class InputParameterOld:
    """Represents an input parameter (old format).

    Attributes
    ----------
    name : str
    value
        The value of this input parameter, usually
        a string, qualified by units
    """

    def __init__(self, raw_data: List) -> None:
        """Initialize InputParameter.

        Parameters
        ----------
        raw_data : List
            Input parameter data as a list.
        """
        self.name = raw_data[1][1][1][1].strip('"')
        self.value = raw_data[1][2]

    @property
    def units(self) -> str:
        """Get the unit label of a Fluent input parameter.

        Returns
        -------
        str
            Unit label of the Fluent input parameter.
        """
        return str(self.value[-1][-1])

    @property
    def numeric_value(self) -> float:
        """Get the numeric value of a Fluent input parameter.

        Returns
        -------
        float
            Numeric value of the Fluent input parameter.
        """
        return float(self.value[2][1])


class InputParameter:
    """Represents an input parameter.

    Attributes
    ----------
    name : str
    value
        The value of this input parameter, usually
        a string, qualified by units
    """

    def __init__(self, raw_data: Dict[str, str]) -> None:
        """Initialize InputParameter.

        Parameters
        ----------
        raw_data : Dict[str, str]
            Input parameter data as a nested dictionary.
        """
        self.name, self.value = None, None
        for k, v in raw_data:
            if k == "name":
                self.name = v.strip('"')
            elif k == "definition":
                self.value = v.strip('"')
                if "[" in self.value:
                    sep_index = self.value.index("[")
                    if self.value[sep_index - 1] != " ":
                        self.value = "".join(
                            (self.value[:sep_index], " ", self.value[sep_index:])
                        )

    @property
    def units(self) -> str:
        """Get the unit label of a Fluent input parameter.

        Returns
        -------
        str
            Unit label of the Fluent input parameter.
        """
        return self._component(1).lstrip("[").rstrip("]")

    @property
    def numeric_value(self) -> float:
        """Get the numeric value of a Fluent input parameter.

        Returns
        -------
        float
            Numeric value of the Fluent input parameter.
        """
        return float(self._component(0))

    def _component(self, idx: int):
        try:
            return self.value.split(maxsplit=1)[idx]
        except IndexError:
            return ""


class OutputParameter:
    """Represents an output parameter.

    Attributes
    ----------
    name : str
    """

    def __init__(self, raw_data: list) -> None:
        """Initialize OutputParameter.

        Parameters
        ----------
        raw_data : list
            Output parameter as a nested list.
        """
        parameter = raw_data[1]
        for elem in parameter:
            if len(elem) and elem[0] == "name":
                self.name = elem[1][1].strip('"')
            if len(elem) and elem[0] == "fluent-units":
                self.units = elem[1].strip('"').strip()


class CaseVariable:
    """Provides access to variables defined in the case."""

    def __init__(self, variables: dict, path: str | None = ""):
        """Initialize CaseVariable.

        Parameters
        ----------
        variables : dict
            The variables dictionary.
        path : str
            The path to the variables.
        """
        self._variables = variables
        self._path = path

    def __call__(self, name: str | None = ""):
        if not name:
            error_name = self._path[:-1] if self._path else self._path
            raise RuntimeError(f"Invalid variable {error_name}")
        try:
            return self._variables[name]
        except KeyError:
            raise ValueError(
                allowed_name_error_message(
                    context="config-vars",
                    trial_name=name,
                    allowed_values=list(self._variables),
                )
            )

    def __getattr__(self, name: str):
        for orig, sub in (
            ("__q", "?"),
            ("__dot", "."),
            ("__plus", "+"),
            ("_", "-"),
        ):
            name = name.replace(orig, sub)
        try:
            name = self._path + name
            result = self._variables[name]
            return lambda: result
        except KeyError:
            return CaseVariable(self._variables, name + "/")


class MeshType(Enum):
    """Types of Mesh."""

    SURFACE = "surface"
    VOLUME = "volume"
    UNKNOWN = "unknown"


def _check_h5_extension(file_path):
    """
    Checks if a given file path ends with the '.h5' extension.

    Parameters
    ----------
    file_path: str
        The path to the file.

    Raises
    ------
    ValueError
        If the file path does not end with '.h5'.
    """
    if not Path(file_path).match("*.h5"):
        raise ValueError("Supports `.h5` extension file format only.")


class Mesh:
    """Class to provide data from and information about Fluent mesh files.

    This class is applicable only to HDF5, Fluent's default format for mesh files.
    Fluent writes HDF5 files with an extension,`.h5` and thus files without that extension
    are not supported.
    HDF5 (Hierarchical Data Format version 5) is commonly used for storing large amounts
    of scientific data, including Fluent mesh data.

    Methods
    -------

    get_surface_ids()
        Get a list of surface ids.
    get_surface_names()
        Get a list of surface names.
    get_surface_locs(surface_id)
        Get the min and max location index of surface.
    get_connectivity(surface_id)
        Get the surface connectivity.
    get_vertices(surface_id)
        Get list of vertices of the surface.
    """

    def __init__(self, file_handle):
        """Initialize the object."""
        _check_h5_extension(file_handle.filename)
        self._file_handle = file_handle

    def get_mesh_type(self) -> MeshType:
        """Returns the type of the mesh."""
        try:
            if "cells" in self._file_handle["meshes"]["1"].keys():
                return MeshType.VOLUME
            else:
                return MeshType.SURFACE
        except Exception:
            return MeshType.UNKNOWN

    def get_surface_ids(self) -> list:
        """Returns list of ids of all available surfaces."""
        id_data = self._file_handle["meshes"]["1"]["faces"]["zoneTopology"]["id"]
        return [id_data[i] for i in range(id_data.size)]

    def get_surface_names(self) -> list:
        """Returns list of names of all available surfaces."""
        return (
            self._file_handle["meshes"]["1"]["faces"]["zoneTopology"]["name"][0]
            .decode()
            .split(";")
        )

    def get_surface_locs(self, surface_id) -> list:
        """Returns range of surface locations for a particular surface."""
        ids = self.get_surface_ids()
        index = ids.index(surface_id)
        min_id = self._file_handle["meshes"]["1"]["faces"]["zoneTopology"]["minId"][
            index
        ]
        max_id = self._file_handle["meshes"]["1"]["faces"]["zoneTopology"]["maxId"][
            index
        ]
        return [int(min_id - 1), int(max_id - 1)]

    def _get_nodes(self, surface_id):
        min_id, max_id = self.get_surface_locs(surface_id)
        nnodes = self._file_handle["meshes"]["1"]["faces"]["nodes"]["1"]["nnodes"]
        nodes = self._file_handle["meshes"]["1"]["faces"]["nodes"]["1"]["nodes"]
        previous = sum(nnodes[0:min_id])
        nnodes = nnodes[min_id : max_id + 1]
        nodes = nodes[previous : previous + sum(nnodes)]
        return [nodes, nnodes]

    def get_connectivity(self, surface_id) -> np.array:
        """Returns numpy array of face connectivity data for a particular surface."""
        nodes, nnodes = self._get_nodes(surface_id)
        key = nodes.copy()
        key.sort()
        key = np.unique(key)
        value = np.arange(0, len(key))
        replace = np.array([key, value])
        mask = np.in1d(nodes, key)
        nodes[mask] = replace[1, np.searchsorted(replace[0, :], nodes[mask])]
        obj = np.cumsum(nnodes)
        obj = np.insert(obj, 0, 0)
        obj = np.delete(obj, len(obj) - 1)
        nodes = np.insert(nodes, obj, nnodes)
        return nodes

    def get_vertices(self, surface_id) -> np.array:
        """Returns numpy array of vertices data for a particular surface."""
        nodes, nnodes = self._get_nodes(surface_id)
        nodes = np.unique(nodes)
        nodes = np.sort(nodes)
        nodes -= 1
        vertices_dict = self._file_handle["meshes"]["1"]["nodes"]["coords"]
        vertices = vertices_dict[str(list(vertices_dict.keys())[0])]
        return vertices[:][nodes].flatten()


class RPVarProcessor:
    """Class to process RP Vars string to expose required outputs.

    Methods
    -------
    input_parameters()
        Get a list of input parameter objects
    output_parameters()
        Get a list of output parameter objects
    num_dimensions()
        Get the dimensionality of the case (2 or 3)
    precision()
        Get the precision (1 or 2 for 1D of 2D)
    iter_count()
        Get the number of iterations
    rp_vars()
        Get dictionary of all RP vars
    rp_var(name)
        Get specific RP var by name, either by providing
        the Scheme name:
            `reader.rp_var("rad/enable-netm?")`
        or a pythonic version:
            `reader.rp_var.rad.enable_netm__q()`
    has_rp_var(name)
        Whether case has particular RP var
    config_vars()
        Get dictionary of all RP vars
    config_var(name)
        Get specific config var by name, either by providing
        the Scheme name:
            `reader.config_var("rp-3d?")`
        or a pythonic version:
            `reader.config_var.rp_3d__q()`
    has_config_var(name)
        Whether case has particular config var
    """

    def __init__(
        self,
        rp_vars_str: str,
    ) -> None:
        """Initialize a RPVarProcessor object.

        Parameters
        ----------
        rp_vars_str :str
            RP Vars string.
        """

        self.rp_vars_str = rp_vars_str

        self._rp_vars = {v[0]: v[1] for v in lispy.parse(rp_vars_str)[1]}

        self._config_vars = {v[0]: v[1] for v in self._rp_vars["case-config"]}

    def input_parameters(self) -> List[InputParameter] | List[InputParameterOld]:
        """Get the input parameters.

        Returns
        -------
        List[InputParameter] | List[InputParameterOld]
            The list of input parameters.
        """
        exprs = self._named_expressions()
        if exprs:
            input_params = []
            for expr in exprs:
                for attr in expr:
                    if attr[0] in ["parameter", "input-parameter"] and attr[1] is True:
                        input_params.append(InputParameter(expr))
            return input_params

        rp_var_params = self._find_rp_var("parameters/input-parameters") or []
        try:
            return [InputParameter(param) for param in rp_var_params]
        except ValueError:
            return [InputParameterOld(param) for param in rp_var_params]

    def output_parameters(self) -> List[OutputParameter]:
        """Get the output parameters.

        Returns
        -------
        List[OutputParameter]
            The list of output parameters.
        """
        parameters = self._find_rp_var("parameters/output-parameters")
        return [OutputParameter(param) for param in parameters]

    def num_dimensions(self) -> int:
        """Get the dimensionality associated with this case.

        Returns
        -------
        int
            The number of dimensions.
        """
        for attr in self._case_config():
            if attr[0] == "rp-3d?":
                return 3 if attr[1] is True else 2

    def precision(self) -> int:
        """Get the precision associated with this case (single or double).

        Returns
        -------
        int
            Either 1 or 2 to indicate single or double precision respectively.
        """
        for attr in self._case_config():
            if attr[0] == "rp-double?":
                return 2 if attr[1] is True else 1

    def iter_count(self) -> int:
        """Get the number of iterations associated with this case.

        Returns
        -------
        int
            The number of iterations associated with this case.
        """
        return self._find_rp_var("number-of-iterations")

    def rp_vars(self) -> dict:
        """Get the rpvars associated with this case.

        Returns
        -------
        dict
            The rpvars associated with this case.
        """
        return self._rp_vars

    @property
    def rp_var(self) -> CaseVariable:
        """Access the rpvars associated with this case.

        Returns
        -------
        CaseVariable
            The rpvars associated with this case.
        """
        return CaseVariable(self._rp_vars)

    def has_rp_var(self, name: str) -> bool:
        """Find if this case has the given rpvar.

        Parameters
        ----------
        name : str
            Name of the rpvar.

        Returns
        -------
        bool
            Whether this case has the given rpvar.
        """
        return name in self._rp_vars

    def config_vars(self) -> dict:
        """Get the config variables associated with this case.

        Returns
        -------
        dict
            The config variables associated with this case.
        """
        return self._config_vars

    @property
    def config_var(self) -> CaseVariable:
        """Access the config variables associated with this case.

        Returns
        -------
        CaseVariable
            The config variables associated with this case.
        """
        return CaseVariable(self._config_vars)

    def has_config_var(self, name):
        """Get whether the case has a given variable."""
        return name in self._config_vars

    def _named_expressions(self):
        return self._find_rp_var("named-expressions")

    def _case_config(self):
        return self._find_rp_var("case-config")

    def _find_rp_var(self, name: str):
        return self._rp_vars.get(name)


class SettingsFile(RPVarProcessor):
    """Class to read a Fluent Settings file."""

    def __init__(self, settings_file_name: str | None = None) -> None:
        """Initialize a SettingsFile object. Exactly one file path argument must be
        specified.

        Parameters
        ----------
        settings_file_name : str
            The path of a settings file.
        """
        if settings_file_name:
            try:
                with open(settings_file_name, "r") as file:
                    rp_vars_str = file.read()
                if not rp_vars_str.startswith("(rp ("):
                    raise RuntimeError("Not a valid settings file.")

            except FileNotFoundError as e:
                raise FileNotFoundError(
                    f"The settings file {settings_file_name} cannot be found."
                ) from e

            except OSError as e:
                raise OSError(
                    f"Error while reading settings file {settings_file_name}"
                ) from e

            except Exception as e:
                raise RuntimeError(
                    f"Could not read settings file {settings_file_name}"
                ) from e

        super().__init__(rp_vars_str)


class EmptyContainer:
    """Empty Container."""

    def __getattr__(self, item):
        return lambda *args, **kwargs: None

    def __call__(self, *args, **kwargs):
        return None


class CaseFile(RPVarProcessor):
    """Class to read a Fluent case file.

    Methods
    -------
    get_mesh()
        Get the mesh data.
    """

    def __init__(
        self,
        case_file_name: str | None = None,
        project_file_name: str | None = None,
    ) -> None:
        """Initialize a CaseFile object. Exactly one file path argument must be
        specified.

        Parameters
        ----------
        case_file_name : str
            The path of a case file.
        project_file_name : str
            The path of a project file from which the case file is selected.
        """
        self._is_case_file = False

        if (not case_file_name) == (not project_file_name):
            raise RuntimeError(
                "Please enter either the case file path or the project file path"
            )
        if project_file_name:
            if Path(project_file_name).suffix in [".flprj", ".flprz"]:
                project_dir = os.path.join(
                    dirname(project_file_name),
                    Path(project_file_name).name.split(".")[0] + ".cffdb",
                )
                case_file_name = Path(
                    project_dir + _get_case_file_name_from_flprj(project_file_name)
                )
            else:
                raise FileNotFoundError(
                    "Please provide a valid fluent project file path"
                )

        try:
            if Path(case_file_name).match("*.cas.h5") or Path(case_file_name).match(
                "*.msh.h5"
            ):
                _file = h5py.File(case_file_name)
                if Path(case_file_name).match("*.h5"):
                    if Path(case_file_name).match("*.cas.h5"):
                        self._is_case_file = True
                        settings = _file["settings"]
                        rpvars = settings["Rampant Variables"][0]
                        rp_vars_str = rpvars.decode()
                    self._mesh = Mesh(_file)
            elif Path(case_file_name).match("*.cas") or Path(case_file_name).match(
                "*.msh"
            ):
                with open(case_file_name, "rb") as _file:
                    rp_vars_str = _file.read()
                if Path(case_file_name).match("*.cas"):
                    self._is_case_file = True
                    rp_vars_str = _get_processed_string(rp_vars_str)
            elif Path(case_file_name).match("*.cas.gz") or Path(case_file_name).match(
                "*.msh.gz"
            ):
                with gzip.open(case_file_name, "rb") as _file:
                    rp_vars_str = _file.read()
                if Path(case_file_name).match("*.cas.gz"):
                    self._is_case_file = True
                    rp_vars_str = _get_processed_string(rp_vars_str)
            else:
                error_message = (
                    "Could not read case file. "
                    "Only valid Case files (.h5, .cas, .cas.gz) or Mesh files (.msh.h5, .msh, .msh.gz) can be read. "
                )
                raise RuntimeError(error_message)

        except FileNotFoundError as e:
            raise FileNotFoundError(
                f"The case file {case_file_name} cannot be found."
            ) from e

        except OSError as e:
            raise OSError(f"Error while reading case file {case_file_name}") from e

        except Exception as e:
            raise RuntimeError(f"Could not read case file {case_file_name}") from e

        if self._is_case_file:
            super().__init__(rp_vars_str=rp_vars_str)

    def get_mesh(self):
        """Get the mesh data."""
        return self._mesh

    def __getattribute__(self, item):
        if (
            item != "_is_case_file"
            and not self._is_case_file
            and item
            in set(filter(lambda k: not k.startswith("__"), dir(RPVarProcessor)))
        ):
            return EmptyContainer()
        return super().__getattribute__(item)


def _get_processed_string(input_string: bytes) -> str:
    """Processes the input string (binary) with help of an identifier to return it in a
    format which can be parsed by lispy.parse().

    Parameters
    ----------
    input_string : bytes
        The input string in bytes

    Returns
    -------
    processed string (str)
    """
    rp_vars_str = codecs.decode(input_string, errors="ignore")
    string_identifier = "(37 ("
    return string_identifier + rp_vars_str.split(string_identifier)[1]


def _get_case_file_name_from_flprj(flprj_file):
    with open(flprj_file, "r") as file:
        content = file.read()
        content = remove_unsupported_xml_chars(content)
        root = ET.fromstring(content)
        folder_name = root.find("Metadata").find("CurrentSimulation").get("value")[5:-1]
        # If the project file name begins with a digit then the node to find will be prepended
        # with "_". Rather than making any assumptions that this is a hard rule, or what
        # the scope of the rule is, simply retry with the name prepended:
        folder_obj = (
            root.find(folder_name)
            if root.find(folder_name) and len(root.find(folder_name)) > 0
            else root.find("_" + folder_name)
        )
        return folder_obj.find("Input").find("Case").find("Target").get("value")



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time

import pytest


@pytest.mark.fluent_version(">=23.1")
@pytest.mark.parametrize(
    "error_code,raises",
    [
        (0, pytest.wont_raise()),
        (1, pytest.raises(RuntimeError)),
    ],
)
def test_fluent_fatal_error(error_code, raises, new_solver_session):
    scheme_eval = new_solver_session.scheme_eval.scheme_eval
    with raises:
        scheme_eval(
            "(events/transmit 'error-event "
            f'(cons (format #f "fatal error: ~a~%" "testing") {error_code}))'
        )
        for _ in range(10):
            # as these are mostly instant, exception should usually be raised on the second gRPC call
            scheme_eval("(pp 'fatal_error_testing)")
            time.sleep(0.1)


@pytest.mark.fluent_version(">=25.2")
def test_custom_python_error_via_grpc(datamodel_api_version_new, new_solver_session):
    solver = new_solver_session
    # This may need to be updated if the error type changes in the server
    with pytest.raises(RuntimeError, match="prefereces not found!"):
        solver._se_service.get_state("prefereces", "General")
    with pytest.raises(ValueError, match="Datamodel rules for prefereces not found!"):
        solver._se_service.get_specs("prefereces", "General")



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from ansys.fluent.core import examples
from ansys.fluent.core.utils.execution import timeout_loop


@pytest.mark.fluent_version(">=23.2")
def test_solver_monitors(new_solver_session):

    solver = new_solver_session

    import_case = examples.download_file(
        file_name="exhaust_system.cas.h5", directory="pyfluent/exhaust_system"
    )

    solver.file.read_case(file_name=import_case)

    ordered_report_plot_names = [
        "mass-bal-rplot",
        "mass-in-rplot",
        "mass-tot-rplot",
        "point-vel-rplot",
    ]

    assert (
        sorted(solver.settings.solution.monitor.report_plots())
        == ordered_report_plot_names
    )

    # monitor set names unavailable without data
    assert len(solver.monitors.get_monitor_set_names()) == 0

    import_data = examples.download_file(
        file_name="exhaust_system.dat.h5", directory="pyfluent/exhaust_system"
    )

    solver.file.read_data(file_name=import_data)

    # monitor set names remains unavailable after loading data
    assert len(solver.monitors.get_monitor_set_names()) == 0

    # monitor set names becomes available after initializing
    solver.solution.initialization.hybrid_initialize()

    monitor_set_names = ordered_report_plot_names + ["residual"]
    assert sorted(solver.monitors.get_monitor_set_names()) == sorted(monitor_set_names)

    # no data in monitors at this point
    assert all(
        all(
            len(elem) == 0
            for elem in solver.monitors.get_monitor_set_data(monitor_set_name=name)
        )
        for name in monitor_set_names
    ), "One or more monitor sets contain non-empty elements."

    # run the solver...
    solver.solution.run_calculation.iterate(iter_count=1)

    # ...data is in monitors
    assert all(
        all(
            len(elem) != 0
            for elem in solver.monitors.get_monitor_set_data(monitor_set_name=name)
        )
        for name in monitor_set_names
    ), "One or more monitor sets contain empty elements."

    def monitor_callback():
        monitor_callback.called = True

    monitor_callback.called = False

    # n.b. there is no checking of the callback signature at registration. Instead
    # we would get a TypeError at callback time if the signature is wrong. The correct
    # signature is undocumented.
    solver.monitors.register_callback(monitor_callback)

    # trigger callback by running the solver
    assert not monitor_callback.called
    solver.solution.run_calculation.iterate(iter_count=1)
    assert timeout_loop(lambda: monitor_callback.called, 5)
    assert monitor_callback.called
